import torch
import numpy as np
import pandas as pd
import torch.nn as nn 
from tcrpeg.model import *      
import argparse
from tcrpeg.utils import *
#from TCRpeg import tcrpeg
import scipy.stats as stats
from tqdm import tqdm

class evaluation:
    def __init__(self,model,vj_model=False):
        '''
        Used for evaluate the probability inference
        @model: the trained TCRpeg model
        @vj_model: whether the model is TCRpeg or TCRpeg_vj
        '''

        self.model = model
        # if path_to_test is not None:
        #     if not vj_model:
        #         self.seqs_test = pd.read_csv(path_to_test,sep='\t').sample(frac=frac)['seq'].values
        #     else :
        #         seqs_test_ = pd.read_csv(path_to_test,sep='\t').sample(frac=frac)
        #         self.seqs_test = seqs_test_['seq'].values
        #         self.vs_test = seqs_test_['v'].values
        #         self.js_test = seqs_test_['j'].values
        self.vj = vj_model
    
    def generate(self,num,batch_size):
        '''
        Generate some new Seqs
        @num: number of generated seqs
        @batch_size: batch_size used in generation

        @return: a list of seqs
        '''
        if self.vj:
            return self.model.generate_tcrpeg_vj( num, batch_size)
        else :
            return self.model.generate_tcrpeg(num,batch_size)

    def eva_prob(self,path,whole=False):
        #if whole=True, will return the r of the whole test set
        #sample 1e6 seqs for r estimation
        #path can be a csv file or a dict with keys of 'seq' and 'count'
        if isinstance(path,dict):
            seq_data = path
        else :
            if not whole:
                seq_data = pd.read_csv(path,compression='gzip').sample(n=int(1e6))
            else :
                seq_data = pd.read_csv(path,compression='gzip')

        c_data_ = list(seq_data['count'])
        seqs_ = list(seq_data['seq'])
        c_data,seqs = [],[]

        for i in range(len(seqs_)):  #only need seqs that has appearance > 2
            if c_data_[i] > 2:
                c_data.append(c_data_[i])
                seqs.append(seqs_[i])
        p_data = np.array(c_data)
        sum_p = np.sum(p_data)
        p_data = p_data / sum_p #normalized probability

        batch_size = 2000

        record = np.zeros(len(seqs))
        with torch.no_grad():        
            for i in tqdm(range(int(len(seqs)/batch_size)+1)):
                end = len(seqs) if (i+1) * batch_size > len(seqs) else (i+1) * batch_size
                seq_batch = seqs[i * batch_size : end]                
                log_probs = self.model.sampling_tcrpeg(seq_batch) #change here
                record[i*batch_size : end] = np.exp(log_probs)
        record_sum = np.sum(record)
        record = record/record_sum
        # kl = kl_divergence(p_data,record)
        corr = stats.pearsonr(p_data,record)[0]
        print('Pearson correlation coefficient are : {}'.format(str(round(corr,4))))
        return corr,p_data,record



from tcrpeg.TCRpeg import TCRpeg
from tcrpeg.classification import classification
import pandas as pd
import numpy as np
import os
from sklearn.metrics import roc_auc_score
from sklearn.metrics import average_precision_score as AUPRC


# g_model = TCRpeg(hidden_size=128,num_layers = 3,load_data=False,embedding_path='data/embedding_32.txt',dropout=0.1)
# g_model.create_model(load=True,path='results/aug/14:32:04/model.pth')
# for num_gens in [2000,6000,10000,14000,18000,25000,30000]:
#     gens = g_model.generate_tcrpeg(num_gens,1000)
#     gens = [x for x in gens if classification.valid_seq(x)]

#     tumor_train = pd.read_csv('data/classification/tumor_train_caTCR.txt',names=['seq'])['seq'].values
#     tumor_test = pd.read_csv('data/classification/tumor_test_caTCR.txt',names=['seq'])['seq'].values
#     normal_train = pd.read_csv('data/classification/normal_train_caTCR.txt',names=['seq'])['seq'].values
#     normal_test = pd.read_csv('data/classification/normal_test_caTCR.txt',names=['seq'])['seq'].values
#     xtrain,ytrain = list(tumor_train)+list(normal_train), [1]*len(tumor_train) + [0] * len(normal_train)
#     xtest,ytest = list(tumor_test)+list(normal_test), [1]*len(tumor_test) + [0] * len(normal_test)
#     xtrain  = xtrain + list(gens)
#     ytrain = ytrain + [1] * len(gens)

#     t_model = TCRpeg(hidden_size=128,num_layers = 3,load_data=False,embedding_path='data/embedding_32.txt')
#     t_model.create_m2oModel()
#     t_model.train_many2one(xtrain,ytrain,25, 8, 1e-3, info=None, model_name=None,record_dir=None)
#     y_pres_whole,ytest = t_model.evaluate_many2one(xtest,ytest,100)
#     print(num_gens)
#     print(roc_auc_score(ytest,y_pres_whole))



np.random.seed(0)
time=0
data = pd.read_csv('../../nlp_vae/data/sars2/cdrs.csv')
cdrs = data['seq'].values
np.random.shuffle(cdrs)
# train = np.array(cdrs[:700])
# test = np.array(cdrs[700:])
test = np.array(cdrs[time*137:(time+1)*137])
train = np.array(list(set(list(cdrs)) - set(list(test))))
data_control = pd.read_csv('../../nlp_vae/data/sars2/control.csv')
cdrs_control = data_control['seq'].values
np.random.shuffle(cdrs_control)
# train_control = cdrs_control[:7000]
# test_control = cdrs_control[7000:]
test_control = np.array(cdrs_control[time*1370:(time+1)*1370])
train_control = np.array(list(set(list(cdrs_control)) - set(list(test_control))))
#train_whole = list(cdrs) + list(cdrs_control)
train_whole = list(train) + list(train_control)
test_whole = list(test) + list(test_control)

t_model = TCRpeg(hidden_size=32,num_layers = 3,load_data=False,embedding_path='data/embedding_32.txt',dropout=0.4)
t_model.create_m2oModel()
t_model.train_many2one(train_whole,[1]*len(train) + [0] * len(train_control),20, 8, 1e-3, info=None, model_name=None,record_dir=None)
y_pres_whole,ytest = t_model.evaluate_many2one(test_whole,[1] * len(test) + [0] * len(test_control),10)
print(roc_auc_score(ytest,y_pres_whole))
print(AUPRC(ytest,y_pres_whole))


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import os

# dirs = os.listdir('results/motifs/')
# d1 = ['results/motifs/' + d for d in dirs if '_1_' in d]
# d2 = ['results/motifs/'+d for d in dirs if 'means_1_' in d]

# motif1 = list('CASQDLNTGELFF'[1:-1])
# motif1_num = list(range(len(motif1)))

# a1 = []
# for i in range(5):
#     print(d1[i])
#     d = pd.read_csv('results/motifs/_1_{}.txt'.format(str(i)),names=['values'])['values'].values
#     a1.append(d)

# a2 = []
# for i in range(5):
#     d = pd.read_csv('results/motifs/means_1_{}.txt'.format(str(i)),names=['values'])['values'].values
#     a2.append(d)

# fig = plt.figure(figsize=(10,5))
# print(len(np.std(a1,0)))
# print(len(motif1_num))
# plt.errorbar(motif1_num,np.mean(a1,0),yerr=np.std(a1,0),fmt='-o',color='b',label='TCRpeg-c',capsize=5)
# plt.errorbar(motif1_num,np.mean(a2,0),yerr=np.std(a2,0),fmt='-o',color='r',label='TCRGP',capsize=5)
# plt.xticks(motif1_num,motif1,size=12)
# plt.ylabel('Prediction score',size=15)
# plt.legend(loc='lower left',fontsize=15)
# plt.tight_layout()
# plt.savefig('results/pictures/motif1.jpg',dpi=200)


# # d1 = ['results/motifs/' + d for d in dirs if '_5_' in d]
# # print(d1)
# # d2 = ['results/motifs/'+d for d in dirs if 'means_5_' in d]

# motif1 = list('CASSPDIEAFF'[1:-1])
# motif1_num = list(range(len(motif1)))

# a1 = []
# for i in range(5):
#     d = pd.read_csv('results/motifs/_5_{}.txt'.format(str(i)),names=['values'])['values'].values
#     a1.append(d)

# a2 = []
# for i in range(5):
#     d = pd.read_csv('results/motifs/means_5_{}.txt'.format(str(i)),names=['values'])['values'].values
#     a2.append(d)

# fig = plt.figure(figsize=(10,5))
# print(len(np.std(a1,0)))
# print(len(motif1_num))
# plt.errorbar(motif1_num,np.mean(a1,0),yerr=np.std(a1,0),fmt='-o',color='b',label='TCRpeg-c',capsize=5)
# plt.errorbar(motif1_num,np.mean(a2,0),yerr=np.std(a2,0),fmt='-o',color='r',label='TCRGP',capsize=5)
# plt.xticks(motif1_num,motif1,size=12)
# plt.ylabel('Prediction score',size=15)
# plt.legend(loc='lower left',fontsize=15)
# plt.tight_layout()
# plt.savefig('results/pictures/motif5.jpg',dpi=200)


###new motif
motif1 = list('CSAKDRGSQETQYF'[1:-1])
motif1_num = list(range(len(motif1)))
a1 = []
for i in range(5):
    d = pd.read_csv('results/motifs/_ETQYF_{}.txt'.format(str(i)),names=['values'])['values'].values
    a1.append(d)
fig = plt.figure(figsize=(10,5))
print(len(np.std(a1,0)))
print(len(motif1_num))
plt.errorbar(motif1_num,np.mean(a1,0),yerr=np.std(a1,0),fmt='-o',color='b',label='TCRpeg-c',capsize=5)
plt.xticks(motif1_num,motif1,size=12)
plt.ylabel('Prediction score',size=15)
#plt.legend(loc='lower left',fontsize=15)
plt.tight_layout()
plt.savefig('results/pictures/motif_ETQYF.jpg',dpi=200)

motif1 = list('CSARDFLGGYTF'[1:-1])
motif1_num = list(range(len(motif1)))
a1 = []
for i in range(5):
    d = pd.read_csv('results/motifs/_GYTF_{}.txt'.format(str(i)),names=['values'])['values'].values
    a1.append(d)
fig = plt.figure(figsize=(10,5))
print(len(np.std(a1,0)))
print(len(motif1_num))
plt.errorbar(motif1_num,np.mean(a1,0),yerr=np.std(a1,0)/np.sqrt(5),fmt='-o',color='b',label='TCRpeg-c',capsize=5)
plt.xticks(motif1_num,motif1,size=12)
plt.ylabel('Prediction score',size=15)
#plt.legend(loc='lower left',fontsize=15)
plt.tight_layout()
plt.savefig('results/pictures/motif_GYTF.jpg',dpi=200)



import numpy as np
from collections import defaultdict
import matplotlib.pyplot as plt
import torch.nn as nn
import torch.nn.functional as F
import matplotlib
from matplotlib import ticker
import pandas as pd
import seaborn as sns
from sklearn.metrics import auc
from sklearn.metrics import roc_curve
from sklearn.metrics import roc_auc_score
from sklearn.metrics import precision_recall_curve as pccurve
from sklearn.metrics import average_precision_score as AUPRC
from scipy.stats import pearsonr as pr

def kl_divergence(dis1,dis2,avoid_zeros=True):
    assert np.abs(np.sum(dis1)-1) <=0.01, 'you need to normalize' 
    if avoid_zeros:
        dis2 = [d + 1e-20 for d in dis2] #to avoid zero division
        dis1 = [d + 1e-20 for d in dis1]
    return sum([dis1[i]*np.log(dis1[i]/dis2[i]) for i in range(len(dis1))])

def js_divergence(p11,p12,p21,p22):
    if np.abs(np.sum(p11) - 1)>0.01:
        p11,p12,p21,p22 = p11/np.sum(p11),p12/np.sum(p12),p21/np.sum(p21),p22/np.sum(p22)
    assert len(p11) == len(p21)
    assert len(p12) == len(p22)
    #p11,p12,p21,p22 = np.array(p11),np.array(p12),np.array(p21),np.array(p22)
    return 0.5*kl_divergence(p11,0.5*(p11+p21)) + 0.5*kl_divergence(p22,0.5*(p22+p12))

class plotting:
    def __init__(self):
        self.aas = 'ACDEFGHIKLMNPQRSTVWY'
        self.aa2idx = {self.aas[i]:i for i in range(len(self.aas))}
        self.idx2aa = {v: k for k, v in self.aa2idx.items()}
        self.vs = ['TRBV10-1','TRBV10-2','TRBV10-3','TRBV11-1','TRBV11-2','TRBV11-3','TRBV12-5', 'TRBV13', 'TRBV14', 'TRBV15', 
                'TRBV16', 'TRBV18','TRBV19','TRBV2', 'TRBV20-1', 'TRBV25-1', 'TRBV27', 'TRBV28', 'TRBV29-1', 'TRBV3-1', 'TRBV30',
        'TRBV4-1', 'TRBV4-2','TRBV4-3',
        'TRBV5-1', 'TRBV5-4', 'TRBV5-5', 'TRBV5-6', 'TRBV5-8','TRBV6-1', 'TRBV6-4', 'TRBV6-5', 'TRBV6-6','TRBV6-8', 'TRBV6-9', 'TRBV7-2', 'TRBV7-3',
        'TRBV7-4', 'TRBV7-6', 'TRBV7-7', 'TRBV7-8', 'TRBV7-9', 'TRBV9']
        self.js = ['TRBJ1-1', 'TRBJ1-2', 'TRBJ1-3','TRBJ1-4', 'TRBJ1-5', 'TRBJ1-6','TRBJ2-1', 'TRBJ2-2', 'TRBJ2-3', 'TRBJ2-4', 'TRBJ2-6', 'TRBJ2-7']


    def valid_seq(self,seq):
        if type(seq) is not str:
            return False
        if len(seq) <= 2:
            return False
        if seq[-1] != 'F' and seq[-2:] != 'YV':

            return False
        if seq[0] != 'C':

            return False
        if 's' in seq or 'e' in seq:
            return False
        return True

    def length_dis(self,data,tcrpeg,sonnia,tcrvae,lengths,fig_name=None):
        '''
        each data input should be in a format (mean_freq,err)
        mean_freq is mapped to lengths
        '''
        plt.figure(figsize=(6,4))
        plt.plot(lengths,data,'b-',linewidth=2,label='Data')
        plt.plot(lengths,tcrpeg[0],'r--',linewidth=1,label='TCRpeg',alpha=0.8)
        plt.fill_between(lengths, tcrpeg[0]-tcrpeg[1], tcrpeg[0]+tcrpeg[1],alpha=0.4)
        plt.plot(lengths,sonnia[0],'g--',linewidth=1,label='soNNia',alpha=0.8)
        #plt.errorbar(lengths, sonnia[0], sonnia[1], fmt='o-', linewidth=1, capsize=5,label='soNNia',markersize=5)
        plt.fill_between(lengths, sonnia[0]-sonnia[1], sonnia[0]+sonnia[1],alpha=0.4)
        plt.plot(lengths,tcrvae[0],'y--',linewidth=1,label='TCRvae',alpha=1)
        #plt.errorbar(lengths, tcrvae[0], tcrvae[1], fmt='o-', linewidth=1, capsize=5,label='TCRvae',markersize=5)
        plt.fill_between(lengths, tcrvae[0]-tcrvae[1], tcrvae[0]+tcrvae[1],alpha=0.4)

        # plt.plot(np.arange(l_length),self.sonia_model.model_marginals[:l_length],label='POST marginals',alpha=0.9)
        # plt.plot(np.arange(l_length),self.sonia_model.data_marginals[:l_length],label='DATA marginals',alpha=0.9)
        # plt.plot(np.arange(l_length),self.sonia_model.gen_marginals[:l_length],label='GEN marginals',alpha=0.9)
        #kl1,kl2,kl3 = kl_divergence(data,tcrpeg[0]),kl_divergence(data,sonnia[0]),kl_divergence(data,tcrvae[0])
        # plt.xticks(rotation='vertical')
        plt.grid()
        plt.legend()
        #plt.title('Length Distribution')
        plt.xlabel('CDR3 length',size=12)
        plt.ylabel('Observed frequency',size=12)
        plt.tight_layout()
        # plt.title('CDR3 LENGTH DISTRIBUTION',fontsize=20)
        if fig_name is not None:
            plt.savefig('results/pictures/'+fig_name + '.jpg',dpi=200)
        else: plt.show()
        #return kl1,kl2,kl3
    def aas_dis(self,data,tcrpeg,sonnia,tcrvae,fig_name=None):
        '''
        inputs should be a dict with len(keys)=20
        '''
        fig,axs = plt.subplots(4,5,figsize=(30,20))
        x_axis = list(range(1,31))
        for row in range(4):
            for col in range(5):
                index = row*5 + col
                d,d1,d2,d3,aa = data[index],tcrpeg[index],sonnia[index],tcrvae[index],self.idx2aa[index]
                y,y1,y2,y3 = [d[k] for k in x_axis],[d1[k][0] for k in x_axis],[d2[k][0] for k in x_axis],[d3[k][0] for k in x_axis]
                e,e1,e2,e3 = [d[k] for k in x_axis],[d1[k][1] for k in x_axis],[d2[k][1] for k in x_axis],[d3[k][1] for k in x_axis]
                # sum1,sum2 = sum(y1),sum(y2)
                # if sum1 == 0 or sum2 == 0:
                #     print(sum1)
                #     print(sum2)
                #     print(aa)
                # y1,y2 = [k/sum1 for k in y1],[k/sum2 for k in y2]
                # kl = kl_divergence(y1,y2)
                #axs[row,col].plot(x_axis,y1,x_axis,y2)
                y1,y2,y3 = np.array(y1),np.array(y2),np.array(y3)
                e1,e2,e3 = np.array(e1),np.array(e2),np.array(e3)
                axs[row,col].plot(x_axis,y,'b-',linewidth=3,label='Data',alpha=1)
                axs[row,col].plot(x_axis, y1,'r--' ,  linewidth=2,label='TCRpeg',alpha=0.8)
                axs[row,col].plot(x_axis, y2,'g--', linewidth=2,label='soNNia',alpha=0.8)
                axs[row,col].plot(x_axis, y3, 'y--', linewidth=2,label='TCRvae',alpha=0.8)
                axs[row,col].fill_between(x_axis, y1-e1, y1+e1, color = 'r',alpha=0.4)
                axs[row,col].fill_between(x_axis, y2-e2, y2+e2,color='g' ,alpha=0.4)
                axs[row,col].fill_between(x_axis, y3-e2, y3+e2, color = 'y', alpha=0.4)
                #axs[row,col].errorbar(x_axis, y3, e3, fmt='o-', linewidth=1, capsize=2,label='TCRvae',markersize=2)
                #axs[row,col].set_title('Amino acid: ' +str(aa))
                axs[row,col].tick_params(axis="x", labelsize=15) 
                axs[row,col].tick_params(axis="y", labelsize=15) 
                axs[row,col].legend(fontsize=15)
                axs[row,col].text(0.02, 0.95, aa, ha='center', va='center',transform = axs[row,col].transAxes,size=20,color='red')
                # axs[row,col].set_xlabel('Position',size=8)
                # axs[row,col].set_ylabel('Frequency',size=8)
                #use a common label for aa dis
        # plt.xticks(fontsize=15)
        # plt.yticks(fontsize=15)
        # fig.text(0.5, -0.01, 'common X', ha='center',size=20)
        # fig.text(0.02, 0.5, 'common Y', va='center', size=20,rotation='vertical')    
        fig.tight_layout()
        if fig_name is not None:
            plt.savefig('results/pictures/'+fig_name + '.jpg',dpi=200)
        else: plt.show()
        

    def pos_dis(self,seqs1:list,seqs2:list,fig_name:str):
        #now is for pos2-10
        pos2aa_dic1 = [defaultdict(int) for _ in range(9)]
        pos2aa_dic2 = [defaultdict(int) for _ in range(9)]

        for i in range(len(seqs1)):
            seq1 = seqs1[i]
            if len(seq1) < 10:
                print(seq1)
                continue
            for j in range(9): #j: 0-8
                pos2aa_dic1[j][seq1[j+1]] += 1
        for i in range(len(seqs2)):
            seq2 = seqs2[i]
            if len(seq2) < 10:
                print(seq2)
                continue
            for j in range(9): #j: 0-8
                pos2aa_dic2[j][seq2[j+1]] += 1
        fig,axs = plt.subplots(3,3,figsize=(16,16))
        x_axis = list('ACDEFGHIKLMNPQRSTVWY')
        for row in range(3):
            for col in range(3):
                index = row*3 + col
                d1,d2,pos = pos2aa_dic1[index],pos2aa_dic2[index],index+2
                y1,y2 = [d1[k] for k in x_axis],[d2[k] for k in x_axis]
                sum1,sum2 = sum(y1),sum(y2)
                y1,y2 = [k/sum1 for k in y1],[k/sum2 for k in y2]
                kl = kl_divergence(y1,y2)
                axs[row,col].plot(x_axis,y1,x_axis,y2)
                axs[row,col].set_title('Position: '+str(pos) +' kl: ' +str(kl))
                axs[row,col].legend(['seqs1','seqs2'])
                plt.savefig('results/pictures/'+fig_name + '.png')

    def aa_num_dis(self,seqs1:list,seqs2:list,fig_name:str):
        aas_1 = defaultdict(int)
        aas_2 = defaultdict(int)
        for i in range(len(seqs1)):
            seq1 = seqs1[i]
            if len(seq1) <= 2 or seq1[0] != 'C' or (seq1[-1] != 'F' and seq1[:-2] != 'YV'):
                continue
            for j,aa in enumerate(seq1):
                #if aa == 'e' or 
                aas_1[aa] += 1
        for i in range(len(seqs2)):
            seq2 = seqs2[i]
            for j,aa in enumerate(seq2):
                aas_2[aa] += 1
        x_axis = list(self.aas)
        y1,y2 = [aas_1[aa] for aa in x_axis],[aas_2[aa] for aa in x_axis]
        sum1,sum2 = sum(y1),sum(y2)
        
        y1,y2 = [k/sum1 for k in y1],[k/sum2 for k in y2]
        kl = kl_divergence(y1,y2)
        plt.figure()
        plt.plot(x_axis,y1,x_axis,y2)
        plt.legend(['fre1','fre2'])
        plt.title(str(kl))
        plt.savefig('results/pictures/'+fig_name + '.png')

    def v_dis(self,data,tcrpeg,sonnia,tcrvae,fig_name=None):
        '''
        each input should be in the format of (mean_freqs,errs)
        also should be one-to-one with the gene list
        '''
        # vs_1 = defaultdict(int)
        # vs_2 = defaultdict(int)
        # for i in range(len(seqs1)):
        #     vs_1[seqs1[i]] += 1
        # for s in seqs2:
        #     vs_2[s] += 1
        # x_axis = self.vs
        # y1,y2 = [vs_1[x]/len(seqs1) for x in vs], [vs_2[x]/len(seqs2) for x in vs]
        # kl = kl_divergence(y1,y2)
        # plt.figure()
        # plt.plot(x_axis,y1,'o',x_axis,y2,'o')
        # #plt.plot(x_axis,y1,x_axis,y2)
        # plt.legend(['fre1','fre2'])
        # plt.title(str(kl))
        # plt.savefig('results/pictures/'+fig_name + '.png')
        plt.figure(figsize=(12,4))
        order=np.argsort(data[0])[::-1]
        x_axis = np.array(list(range(len(self.vs))))
        plt.errorbar(x_axis-0.2, np.array(data[0])[order] , fmt='b+', linewidth=4, capsize=5,label='Data',markersize=8)
        plt.errorbar(x_axis-0.1, np.array(tcrpeg[0])[order],np.array(tcrpeg[1])[order] , fmt='ro', linewidth=1, capsize=3,label='TCRpeg',markersize=3)
        plt.errorbar(x_axis+0.1, np.array(sonnia[0])[order] ,np.array(sonnia[1])[order], fmt='go', linewidth=1, capsize=3,label='soNNia',markersize=3)
        plt.errorbar(x_axis+0.2, np.array(tcrvae[0])[order] ,np.array(tcrvae[1])[order], fmt='yo', linewidth=1, capsize=3,label='TCRvae',markersize=3)
        plt.xticks(x_axis,np.array(self.vs)[order])
        plt.xticks(rotation=45)
        plt.ylabel('Frequency',size=12)
        plt.xlabel('V Gene',size=12)
        plt.grid()
        plt.legend()
        #plt.title('J USAGE DISTRIBUTION',fontsize=20)
        plt.tight_layout()
        if fig_name is not None:
            plt.savefig('results/pictures/'+fig_name + '.jpg',dpi=200)
        else: plt.show()
        


    def j_dis(self,data,tcrpeg,sonnia,tcrvae,fig_name=None):
        plt.figure(figsize=(6,4))
        order=np.argsort(data[0])[::-1]
        x_axis = np.array(list(range(len(self.js))))
        plt.errorbar(x_axis-0.2, np.array(data[0])[order] , fmt='b+', linewidth=4, capsize=3,label='Data',markersize=10)
        plt.errorbar(x_axis-0.1, np.array(tcrpeg[0])[order],np.array(tcrpeg[1])[order] , fmt='ro', linewidth=1, capsize=3,label='TCRpeg',markersize=3)
        plt.errorbar(x_axis+0.1, np.array(sonnia[0])[order] ,np.array(sonnia[1])[order], fmt='go', linewidth=1, capsize=3,label='soNNia',markersize=3)
        plt.errorbar(x_axis+0.2, np.array(tcrvae[0])[order] ,np.array(tcrvae[1])[order], fmt='yo', linewidth=1, capsize=3,label='TCRvae',markersize=3)
        plt.xticks(x_axis,np.array(self.js)[order])
        plt.xticks(rotation=45)
        plt.ylabel('Frequency',size=12)
        plt.xlabel('J Gene',size=12)
        plt.grid()
        plt.legend()
        #plt.title('J USAGE DISTRIBUTION',fontsize=20)
        plt.tight_layout()
        if fig_name is not None:
            plt.savefig('results/pictures/'+fig_name + '.jpg',dpi=200)
        else: plt.show()

    def density_scatter(self, x , y, ax = None, method='tcrpeg',sort = True, bins = 20, fig_name=None )   :
        """
        Scatter plot colored by 2d histogram
        """
        if ax is None :
            fig , ax = plt.subplots()
        bins = [100, 500] # number of bins
        # histogram the data
        hh, locx, locy = np.histogram2d(x, y, bins=bins)
        hh=hh#/hh.max()
        z = np.array([hh[np.argmax(a<=locx[1:]),np.argmax(b<=locy[1:])] for a,b in zip(x,y)])
        idx = z.argsort()
        x2, y2, z2 = x[idx], y[idx], z[idx]
        map_reversed = matplotlib.cm.get_cmap('viridis_r')
        s=ax.scatter(x2, y2, c=z2, cmap=map_reversed,s=10,alpha=1.,rasterized=True)
        tick_locator = ticker.MaxNLocator(nbins=4)
        cb=plt.colorbar(s,ax=ax,ticks=tick_locator)
        cb.ax.tick_params(labelsize=5)
        ax.text(0.1, 0.9, method, ha='center', va='center',transform = ax.transAxes,size=10,color='k')
        #plt.tight_layout()
        if fig_name is not None:
            plt.savefig('results/pictures/'+fig_name + '.jpg',dpi=200)
        #else: plt.show()


    def plot_model_learning(self,train,val,label, save_name = None):
        
        """Plots L1 convergence curve
        Parameters
        ----------
        save_name : str or None
            File name to save output figure. If None (default) does not save.
        """

        fig = plt.figure(figsize =(4, 4))
        
        # fig.add_subplot(122)
        # plt.title('Likelihood', fontsize = 15)
        plt.plot(train,label='train',c='k')
        plt.plot(val,label='validation',c='r')
        plt.legend(fontsize = 10)
        plt.xlabel('Iteration', fontsize = 13)
        plt.ylabel(label, fontsize = 13)

        fig.tight_layout()

        if save_name is not None:
            fig.savefig(save_name)
        else: plt.show()

    def plot_violin(self,a1,a2,epitopes,save_name=None,method1='TCRpeg-c',method2 = 'TCRGP' ):
        '''
        a1, a2 should a list of list, each list should contain the metric value for each experiment
        Note that each sub_list should correspond to the epitope in epitopes (the order should be match)
        '''
        a1,a2,epitopes = np.array(a1),np.array(a2),np.array(epitopes) #a1 is 5 x 22
        es,vs,ms = [],[],[]
        means = [-np.mean(x) for x in a1]
        index = np.argsort(means)
        a1,a2,epitopes = a1[index],a2[index],epitopes[index]
        for i in range(len(a1)):
            vs += list(a1[i])
            vs += list(a2[i])
            es += ([epitopes[i]]  * (len(a1[i]) * 2))
            ms += [method1] * len(a1[i])
            ms += [method2]* len(a2[i])
    #         print(len(vs))
    #         print(len(es))
        df = pd.DataFrame({'Epitope':es,'AUPRC':vs,'Method':ms})
        plt.figure(figsize = (20,7))
        sns.violinplot(data = df, y = "AUPRC", x = "Epitope", hue = "Method")
        plt.grid(axis='x')
        plt.xlabel('Epitope',size=12)
        plt.ylabel('AUPRC',size=12)
        plt.xticks(rotation=45)
        plt.tight_layout()
        if save_name is not None:
            plt.savefig(save_name+'.jpg',dpi=200)

    def plot_sup(self,a1,a2,save_name=None ,xlabel='Mean AUPRC of TCRGP',ylabel='Mean AUPRC of TCRpeg-c'):
        '''
        a1, a2 should a list of list, each list should contain the metric value for each experiment 1 is TCRpeg
        Note that each sub_list should correspond to the epitope in epitopes (the order should be match)
        '''
        a1,a2 = np.array(a1),np.array(a2)
        es,vs,ms = [],[],[]
        mean_1,mean_2 = np.mean(a1,1),np.mean(a2,1)
        infer = (mean_1 < mean_2)
        colors = np.array(['#578EB9'] * 22)
        colors[infer] = '#FFA500'
        
        plt.figure(figsize = (6,4.5))
        plt.scatter(mean_2,mean_1,c=colors)
        xpoints = ypoints = plt.xlim()
        plt.plot(xpoints, ypoints, linestyle='--', color='k', lw=2, scalex=False, scaley=False)
        plt.xlabel(xlabel,size=12)
        plt.ylabel(ylabel,size=12)
        if save_name is not None:
            plt.savefig(save_name+'.jpg',dpi=200)

    def plot_auc(self,a1,a2,y_true1,y_true2,name1='TCRpeg-c',name2='DeepCAT',save_name = None):
        '''
        list of list, each list is the prediction scores
        for example, 5 x len(y_true)
        '''
        fig, ax = plt.subplots()
        tprs_1,aucs_1 = [] ,[]
        tprs_2,aucs_2 = [] ,[]
        mean_fpr = np.linspace(0, 1, 1000)
        #scores = classifier.predict_proba(X[test])
        for i in range(len(a1)):
            viz_1 = roc_curve(
                    y_true1[i],
                    a1[i]
                )
            viz_2 = roc_curve(
                    y_true2[i],
                    a2[i]
                )
            interp_tpr = np.interp(mean_fpr, viz_1[0], viz_1[1])
            interp_tpr[0] = 0.0
            tprs_1.append(interp_tpr)
            aucs_1.append(roc_auc_score(y_true1[i],a1[i]))
            interp_tpr = np.interp(mean_fpr, viz_2[0], viz_2[1])
            interp_tpr[0] = 0.0
            tprs_2.append(interp_tpr)
            aucs_2.append(roc_auc_score(y_true2[i],a2[i]))
        print(aucs_1)   
        print(aucs_2)
        # exit()
        ax.plot([0, 1], [0, 1], linestyle="--", lw=2, color="k", alpha=0.8)
        mean_tpr1 = np.mean(tprs_1, axis=0)
        mean_tpr2 = np.mean(tprs_2, axis=0)
        mean_tpr1[-1] = 1.0
        mean_tpr2[-1] = 1.0
        mean_auc1 = auc(mean_fpr, mean_tpr1)
        mean_auc2 = auc(mean_fpr, mean_tpr2)
        std_auc1 = np.std(aucs_1)  
        std_auc2 = np.std(aucs_2) 
        
        
        ax.plot(
            mean_fpr,
            mean_tpr1,
            color="b",
            label=name1+ r" (AUC = %0.3f $\pm$ %0.3f)" % (mean_auc1, std_auc1),
            lw=1.5,
            alpha=0.5,
        )
        ax.plot(
            mean_fpr,
            mean_tpr2,
            color="r",
            label=name2+ r" (AUC = %0.3f $\pm$ %0.3f)" % (mean_auc2, std_auc2),
            lw=1.5,
            alpha=0.8,
        )
        std_tpr1 = np.std(tprs_1, axis=0)
        tprs_upper = np.minimum(mean_tpr1 + std_tpr1, 1)
        tprs_lower = np.maximum(mean_tpr1 - std_tpr1, 0)
        ax.fill_between(
            mean_fpr,
            tprs_lower,
            tprs_upper,
            color="b",
            alpha=0.2
        )
        std_tpr2 = np.std(tprs_2, axis=0)
        tprs_upper = np.minimum(mean_tpr2 + std_tpr2, 1)
        tprs_lower = np.maximum(mean_tpr2 - std_tpr2, 0)
        ax.fill_between(
            mean_fpr,
            tprs_lower,
            tprs_upper,
            color="r",
            alpha=0.2
        )
        plt.legend(loc="lower right")
        plt.xlabel('False Positive Rate',size=12)
        plt.ylabel('True Positive Rate',size=12)
        if save_name is not None:
            plt.savefig(save_name + '.jpg',dpi=200)
        else :
            plt.show()

    def plot_prc(self,a1,a2,y_true1,y_true2,name1='TCRpeg-c',name2='DeepCAT',save_name = None):
        '''
        list of list, each list is the prediction scores
        for example, 5 x len(y_true)
        '''
        fig, ax = plt.subplots()
        tprs_1,aucs_1 = [] ,[]
        tprs_2,aucs_2 = [] ,[]
        mean_fpr = np.linspace(0, 1, 1500)
        #scores = classifier.predict_proba(X[test])
        for i in range(len(a1)):
            p1,r1,_ = pccurve(
                    y_true1[i],
                    a1[i]
                )
            p2,r2,_ = pccurve(
                    y_true2[i],
                    a2[i]
                )
            r1,p1,r2,p2 = r1[::-1],p1[::-1],r2[::-1],p2[::-1]
            interp_tpr = np.interp(mean_fpr, r1, p1)
            print(interp_tpr)
            interp_tpr[0] = 1.0
            tprs_1.append(interp_tpr)
            aucs_1.append(AUPRC(y_true1[i],a1[i]))
            interp_tpr = np.interp(mean_fpr, r2, p2)
            interp_tpr[0] = 1.0
            tprs_2.append(interp_tpr)
            aucs_2.append(AUPRC(y_true2[i],a2[i]))
        #ax.plot([0, 1], [0, 1], linestyle="--", lw=2, color="k", alpha=0.8)
        mean_tpr1 = np.mean(tprs_1, axis=0)
        mean_tpr2 = np.mean(tprs_2, axis=0)
        # mean_tpr1[-1] = 0.0
        # mean_tpr2[-1] = 0.0
        mean_auc1 = auc(mean_fpr, mean_tpr1)
        mean_auc2 = auc(mean_fpr, mean_tpr2)
        std_auc1 = np.std(aucs_1)  
        std_auc2 = np.std(aucs_2) 
        print(aucs_1)
        print(aucs_2)
        
        ax.plot(
            mean_fpr,
            mean_tpr1,
            color="b",
            label=name1+ r" (AUC = %0.3f $\pm$ %0.3f)" % (mean_auc1, std_auc1),
            lw=1.5,
            alpha=0.5,
        )
        ax.plot(
            mean_fpr,
            mean_tpr2,
            color="r",
            label=name2+ r" (AUC = %0.3f $\pm$ %0.3f)" % (mean_auc2, std_auc2),
            lw=1.5,
            alpha=0.8,
        )
        std_tpr1 = np.std(tprs_1, axis=0)
        tprs_upper = np.minimum(mean_tpr1 + std_tpr1, 1)
        tprs_lower = np.maximum(mean_tpr1 - std_tpr1, 0)
        ax.fill_between(
            mean_fpr,
            tprs_lower,
            tprs_upper,
            color="b",
            alpha=0.2
        )
        std_tpr2 = np.std(tprs_2, axis=0)
        tprs_upper = np.minimum(mean_tpr2 + std_tpr2, 1)
        tprs_lower = np.maximum(mean_tpr2 - std_tpr2, 0)
        ax.fill_between(
            mean_fpr,
            tprs_lower,
            tprs_upper,
            color="r",
            alpha=0.2
        )
        plt.legend(loc="lower left")
        plt.xlabel('Recall',size=12)
        plt.ylabel('Precisiion',size=12)
        if save_name is not None:
            plt.savefig(save_name + '.jpg',dpi=200)
        else :
            plt.show()



    def plot_auc_aug(self, inputs,y_true, nums ,save_name=None):
        fig, ax = plt.subplots()
        #originals = [np.mean([auc(y_true,x) for x in input]) for input in inputs]
        mean_fpr = np.linspace(0, 1, 1000)
        for k in range(len(nums)):
            tprs_1,aucs_1 = [] ,[]
            for i in range(len(inputs[k])):
                viz_1 = roc_curve(
                        y_true[k][i],
                        inputs[k][i]
                    )
                interp_tpr = np.interp(mean_fpr, viz_1[0], viz_1[1])
                interp_tpr[0] = 0.0
                tprs_1.append(interp_tpr)
                aucs_1.append(roc_auc_score(y_true[k][i],inputs[k][i]))
                
            ax.plot([0, 1], [0, 1], linestyle="--", lw=2, color="k", alpha=0.8)
            mean_tpr1 = np.mean(tprs_1, axis=0)
            mean_tpr1[-1] = 1.0
            mean_auc1 = auc(mean_fpr, mean_tpr1)
            std_auc1 = np.std(aucs_1)  
            
            if k == len(nums) - 1:
                ax.plot(
                mean_fpr,
                mean_tpr1,
                label=r"No augmentation (AUC = %0.3f $\pm$ %0.3f)" % (0.897, 0.002),
                lw=2,
                alpha=1,
                color='k'
                )
                std_tpr1 = np.std(tprs_1, axis=0)
                tprs_upper = np.minimum(mean_tpr1 + std_tpr1, 1)
                tprs_lower = np.maximum(mean_tpr1 - std_tpr1, 0)
                ax.fill_between(
                    mean_fpr,
                    tprs_lower,
                    tprs_upper,
                    alpha=0.1,
                    color='k'
                )
            else :
                ax.plot(
                    mean_fpr,
                    mean_tpr1,
                    label=r"%d (AUC = %0.3f $\pm$ %0.3f)" % (nums[k],np.mean(aucs_1), std_auc1),
                    lw=1,
                    alpha=0.5,
                )
                std_tpr1 = np.std(tprs_1, axis=0)
                tprs_upper = np.minimum(mean_tpr1 + std_tpr1, 1)
                tprs_lower = np.maximum(mean_tpr1 - std_tpr1, 0)
                ax.fill_between(
                    mean_fpr,
                    tprs_lower,
                    tprs_upper,
                    alpha=0.2
                )
        plt.legend(loc="lower right")
        plt.xlabel('False Positive Rate',size=12)
        plt.ylabel('True Positive Rate',size=12)
        if save_name is not None:
            plt.savefig(save_name + '.jpg',dpi=200)
        else :
            plt.show()

    def augmentation(self,nums,vs,original=None,method_name='Augmentation by TCRpeg',ax=None,save_name=None):
        '''
        original is just a list
        nums is the number of generated seqs; vs is the performances under that num, should be in shape len(nums)x n
        #now is only for TCRpeg
        '''
        if ax is None:
            fig, ax = plt.subplots()
        mean_v = np.mean(vs,axis=1)
        print(mean_v)
        std_v = np.std(vs,axis=1) / np.sqrt(5)
        print(std_v)
        ax.plot(nums,
            mean_v,
            label=method_name,
            lw=2,
            alpha=0.8)
        lower,upper = mean_v + std_v, mean_v - std_v
        ax.fill_between(
            nums,
            lower,
            upper,
            alpha=0.2
        )
        if original is not None:
            mean = np.mean(original)
            mean = np.array([mean]*len(nums))
            std = np.std(original)/5
            lower,upper = mean + std, mean - std
            ax.plot(nums, mean,label='No augmentation',lw=2,alpha=0.8)
            ax.fill_between(nums,lower,upper,alpha=0.2)
            
        plt.xlabel('Number of generated TCR sequences',size=12)
        plt.ylabel('AUROC',size=12)
        plt.legend(loc='upper right')
        if save_name is not None:
            plt.savefig(save_name + '.jpg',dpi=200)
        else :
            plt.show()

    def Length_Dis(self,seqs1:list,seqs2 :list,save_path=None):
        'plot the length distribution for a single model in one experiment'

        lens1 = [seq for seq in seqs1 if self.valid_seq(seq)]
        lens1,lens2 = [len(seq) for seq in seqs1],[len(seq) for seq in seqs2]
        len2count1,len2count2 = defaultdict(int),defaultdict(int)
        for i in range(len(lens1)):        
            len2count1[lens1[i]] += 1
        for i in range(len(lens2)):
            len2count2[lens2[i]] += 1
        k1,k2 = list(len2count1.keys()),list(len2count2.keys())
        x_axis = list(range(1,31))
        fre1,fre2 = [len2count1[k]/len(lens1) for k in x_axis],[len2count2[k]/len(lens2) for k in x_axis]
        plt.figure()
        plt.plot(x_axis,fre1,x_axis,fre2)
        plt.legend(['Data','Generated'])
        plt.xlabel('TCR Length')
        plt.ylabel('Frequency')
        if save_path is not None:
            plt.savefig(save_path+ '.png')
        else :
            plt.show()

    def AAs_Dis(self,seqs1:list,seqs2:list,save_path = None):
        'plot the position distribution for AA, for single model and single experiment'        
        aa2pos_dic1 = [defaultdict(int) for _ in range(20)]
        aa2pos_dic2 = [defaultdict(int) for _ in range(20)]
        for i in range(len(seqs1)):
            seq1 = seqs1[i]
            if not self.valid_seq(seq1):
                continue
            for j,aa in enumerate(seq1):
                #if aa == 'e' or 
                aa2pos_dic1[self.aa2idx[aa]][j+1] += 1
        for i in range(len(seqs2)):
            seq2 = seqs2[i]
            if not self.valid_seq(seq2):
                continue
            for j,aa in enumerate(seq2):
                aa2pos_dic2[self.aa2idx[aa]][j+1] += 1
        
        fig,axs = plt.subplots(4,5,figsize=(20,15))
        x_axis = list(range(1,31))
        for row in range(4):
            for col in range(5):
                index = row*5 + col
                d1,d2,aa = aa2pos_dic1[index],aa2pos_dic2[index],self.idx2aa[index]
                y1,y2 = [d1[k] for k in x_axis],[d2[k] for k in x_axis]
                sum1,sum2 = sum(y1),sum(y2)            
                y1,y2 = [k/sum1 for k in y1],[k/sum2 for k in y2]
                axs[row,col].plot(x_axis,y1,x_axis,y2)

                axs[row,col].legend(['Data','Generated'])
                axs[row,col].set_title('AA: ' +aa)
        plt.tight_layout()
        if save_path is not None:
            plt.savefig(save_path +  '.png')
        else :
            plt.show()

    def plot_prob(self,p_data,p_infer,save_path=None):
        'plot the scatter plot for Fig.1 in the paper'
        fig=plt.figure(figsize=(4,2),dpi=200)
        ax1=plt.subplot(111)
        ax1.set_ylim([-21,0])
        ax1.set_xlim([-5.5,-2.5])
        ax1.locator_params(nbins=4)
        ax1.set_xlabel(r'$log_{10}P_{data}$')
        ax1.set_ylabel(r'$log_{10}P_{infer}$')
        ax1.plot([-5.5, -2.5], [-5.5, -2.5], color='k', linestyle='-', linewidth=2)
        #Plot.density_scatter(np.log10(p_data),np.log10(p_tcrpeg),bins = [10,50],ax=ax1,fig_name='prob_tcrpeg',method='TCRpeg')
        self.density_scatter(np.log10(p_data),np.log10(p_infer),bins = [10,50],ax=ax1,method='TCRpeg')
        r = pr(p_data,p_infer)[0]    
        ax1.text(0.65, 0.32, r'r = %0.3f' % (r) , ha='center', va='center',transform = ax1.transAxes,size=10,color='k')
        plt.tight_layout()
        if save_path is not None:
            plt.savefig(save_path+'.jpg',dpi=200)
        else :
            plt.show()


import pandas as pd
import numpy as np
from utils import *
import os
from collections import defaultdict
from sklearn.metrics import PrecisionRecallDisplay
from sklearn.metrics import precision_recall_curve as pccurve

#plot for auc
dirs = os.listdir('results/caTCR/predictions/')
#dirs = os.listdir('results/sars2/')
dirs_1 = [d for d in dirs if 'auc' in d]
a1 = []
y_trues1 = []
for i in range(5):
    d_peg = pd.read_csv('results/caTCR/predictions/pres_large_2_cross_{}.txt'.format(str(i)),names=['y_pres','y_trues'])
    #d_peg = pd.read_csv('results/sars2/auc_{}.txt'.format(str(i)),names = ['y_trues','y_pres'])
    a1.append(list(d_peg['y_pres'].values))
    y_trues1.append(list(d_peg['y_trues'].values))

# dirs_2 = ['results/caTCR/predictions/' + x for x in dirs if 'deepcat' in x]
# #dirs_2 = [d for d in dirs if 'scores' in d]
# a2= []
# y_trues2 = []
# for d in dirs_2:
#     #d_cat = pd.read_csv('results/sars2/'+ d,names=['y_trues','y_pres'])
#     d_cat = pd.read_csv(d,names=['y_pres','y_trues'])
#     a2.append(list(d_cat['y_pres'].values))
#     y_trues2.append(list(d_cat['y_trues'].values))

# Plot = plotting()
# Plot.plot_auc(a1,a2,y_trues1,y_trues2,name2 = 'DeepCAT',save_name='results/pictures/auc_test_deepcat')


# print(len(y_trues1))
# print(len(y_trues2))
# # Plot.plot_auc(a1,a2,y_trues1,y_trues2,name2 = 'TCRGP')
# Plot.plot_prc(a1,a2,y_trues1,y_trues2,name2 = 'TCRGP')


# display = PrecisionRecallDisplay.from_predictions(y_trues1, d_peg['y_pres'].values, name="LinearSVC")
# display = PrecisionRecallDisplay.from_predictions(y_trues1, d_cat['y_pres'].values, name="JYP")
# plt.xlabel('jyp')
# plt.show()

#Plot.plot_prc(a1,a2,y_trues1,y_trues2,name2='TCRGP',save_name='results/pictures/aup_sars')


# nums = [2000,6000,10000,12000,14000,16000,18000,20000,25000,30000]
# res = defaultdict(list)
# base = 'results/aug_deepcat/'
# for i in range(5):
#     for num in nums:
#         d_cat = pd.read_csv(base + '{}_num_augmented_{}.txt'.format(i+1,num),names=['y_pres','y_trues'])
#         res[num].append(list(d_cat['y_pres'].values))
# y_trues = d_cat['y_trues'].values
# input_ = []
# for num in nums:
#     input_.append(res[num])
# # input_.append(a2)
# # nums.append(0)
# #Plot.plot_auc_aug(input_,y_trues,nums,'results/pictures/auc_aug')
# a2 = [roc_auc_score(y_trues,a) for a in a2]
# input_ = [[roc_auc_score(y_trues,y_pre) for y_pre in temps] for temps in input_]
# Plot.augmentation(nums,input_,a2,save_name='results/pictures/aug_line')


# dirs = os.listdir('results/caTCR/predictions/')
# dirs = [x for x in dirs if 'pres_large' in dirs]
# a2 = []
# y_trues1 = []
# for i in range(5):
#     d_peg = pd.read_csv('results/caTCR/predictions/pres_large_2_cross_{}.txt'.format(str(i)),names=['y_pres','y_trues'])
#     a2.append(list(d_peg['y_pres'].values))
#     y_trues1.append(list(d_peg['y_trues'].values))

Plot = plotting()
# Plot.plot_auc(a1,a2,y_trues1,y_trues2,'results/pictures/auc')
nums = [8000,12000,14000,16000,18000,20000,22000,24000]
file_nums = [int(n/0.4) for n in nums]
res = defaultdict(list)
base = 'results/aug2/predictions/'
res_y = defaultdict(list)
for i in range(5):
    for num in file_nums:
        d_cat = pd.read_csv(base + '{}_pres_large_cross_last_{}.txt'.format(i,num),names=['y_pres','y_trues'])
        res[int(num * 0.4)].append(list(d_cat['y_pres'].values))
        res_y[int(num * 0.4)].append(list(d_cat['y_trues'].values))
input_ = []
input_y = []
for num in nums:
    input_.append(res[num])
    input_y.append(res_y[num])
input_.append(a1)
input_y.append(y_trues1)
nums.append(0)
Plot.plot_auc_aug(input_,input_y,nums)
# a2 = [roc_auc_score(y_trues,a) for a in a2]
# input_ = [[roc_auc_score(y_trues,y_pre) for y_pre in temps] for temps in input_]
# # print(len(input_))
# # print(len(input_[0]))
# Plot.augmentation(nums,input_,a2,save_name='results/pictures/aug_line_tcrpegc')



import pandas as pd
import numpy as np
import os

'''
How to use this code for getting the universal TCR pool from Emerson et al.:
First put the 743 files of individual repertoires under the folder 'data_original/', and then
in terminal, type:
python process_data.py

After execution, it will generate two files 'whole_seqs_train_nn.tsv' and 'whole_seqs_train_nn.tsv' under the 'data/' folder (So you need to create folder 'data' in advance)
'''


vs_default = ['TRBV10-1','TRBV10-2','TRBV10-3','TRBV11-1','TRBV11-2','TRBV11-3','TRBV12-5', 'TRBV13', 'TRBV14', 'TRBV15', 
        'TRBV16', 'TRBV18','TRBV19','TRBV2', 'TRBV20-1', 'TRBV25-1', 'TRBV27', 'TRBV28', 'TRBV29-1', 'TRBV3-1', 'TRBV30',
'TRBV4-1', 'TRBV4-2','TRBV4-3',
 'TRBV5-1', 'TRBV5-4', 'TRBV5-5', 'TRBV5-6', 'TRBV5-8','TRBV6-1', 'TRBV6-4', 'TRBV6-5', 'TRBV6-6','TRBV6-8', 'TRBV6-9', 'TRBV7-2', 'TRBV7-3',
  'TRBV7-4', 'TRBV7-6', 'TRBV7-7', 'TRBV7-8', 'TRBV7-9', 'TRBV9']
js_default = ['TRBJ1-1', 'TRBJ1-2', 'TRBJ1-3','TRBJ1-4', 'TRBJ1-5', 'TRBJ1-6','TRBJ2-1', 'TRBJ2-2', 'TRBJ2-3', 'TRBJ2-4', 'TRBJ2-6', 'TRBJ2-7']

def get_whole_data(files,aa_column_beta=1,v_beta_column=10,j_beta_column=16,frame_column=2,nc_column=0,max_length=30,filter_genes=True):
    #Pool the universal TCR pool from Emerson et al.

    #filter_genes: whether to filter genes that not in vs_default and js_default. This step is just used to follow soNNia and TCRvae since 
    #they all excludes genes that not in these two lists

    beta_sequences = []
    v_beta,j_beta = [],[]
    nn_beta = []
    nn_unpro,v_un,j_un = [],[],[]
    print('loading Data')
    in_frame=True
    seps = ['\t' if f.endswith('.tsv') else ',' for f in files]
    unique_seqs = set() # to store nn now
    count = 0

    for i,file in enumerate(files):
        if i % 10 == 0:
            print('Have processed {} files'.format(i+1))
        f = pd.read_csv(file,sep=seps[i])
        col_nn = f[f.columns[nc_column]].values
        col_v,col_j = f[f.columns[v_beta_column]].values,f[f.columns[j_beta_column]].values
        col_beta = f[f.columns[aa_column_beta]].values if in_frame else f[f.columns[nc_column]]
        frame_bool = f[f.columns[frame_column]].values 
        frames_out = [f for f in frame_bool if f != 'In' or f!='in']
        # if len(frames_out) <= 10000:
        #     count += 1
        #     continue
        for k in range(len(col_beta)):
            if frame_bool[k] != 'In' or frame_bool[k] != 'in':
                if col_v[k] !='unresolved' and col_j[k] != 'unresolved':
                    if type(col_v[k]) is str and type(col_j[k]) is str:
                        if col_j[k] != 'TCRBJ02-05':
                            #nn_unpro.append(col_nn[k])
                            # v_un.append(col_v[k])
                            # j_un.append(col_j[k])
                            pass
        
            if not PassFiltered_((col_beta[k],col_v[k],col_j[k]),frame_bool[k],in_frame,max_length,filter_genes):
                    continue  
            if col_nn[k] in unique_seqs:
                continue
            beta_sequences.append(col_beta[k])
            unique_seqs.add(col_nn[k])
            v_beta.append(col_v[k])
            j_beta.append(col_j[k])
            nn_beta.append(col_nn[k])
        print(len(beta_sequences))
    print(count)
    beta_sequences,v_beta,j_beta,nn_beta = np.array(beta_sequences),np.array(v_beta),np.array(j_beta),np.array(nn_beta)
    infer_index = np.random.permutation(list(range(len(beta_sequences))))
    res_train = pd.DataFrame(columns=['seq','v','j','nn'])
    res_train['seq'] = beta_sequences[infer_index[:int(len(infer_index)/2)]]
    res_train['v'] = v_beta[infer_index[:int(len(infer_index)/2)]]
    res_train['j'] = j_beta[infer_index[:int(len(infer_index)/2)]]
    res_train['nn'] = nn_beta[infer_index[:int(len(infer_index)/2)]]
    res_train.to_csv('data/whole_seqs_train_nn.tsv',sep='\t',index=False,compression='gzip')
    res_test = pd.DataFrame(columns=['seq','v','j','nn'])
    res_test['seq'] = beta_sequences[infer_index[int(len(infer_index)/2):]] 
    res_test['v'] = v_beta[infer_index[int(len(infer_index)/2):]] 
    res_test['j'] = j_beta[infer_index[int(len(infer_index)/2):]]
    res_test['nn'] = nn_beta[infer_index[int(len(infer_index)/2):]]
    res_test.to_csv('data/whole_seqs_test_nn.tsv',sep='\t',index=False,compression='gzip')

def PassFiltered_(to_filter, frame_bool,in_frame,max_length,filter_genes):
    ### to check whether the row of data is valid
    seq,v,j = to_filter
    if in_frame:
        if frame_bool != 'In':
            return False             
    # else :
    #     if frame_bool != 'Out':
    #         return False
    if type(seq) is not str:            
        return False
    if type(v) is not str or type(j) is not str:
        return False
    if len(seq) > max_length or '*' in seq or 'C' != seq[0] or ('F' != seq[-1] and 'YV' != seq[-2:]):
        return False
    if v =='unresolved' or j == 'unresolved':
        return False

    if j == 'TCRBJ02-05' : #based on data preprocessing step in soNNia
        return False
    if seq =='CFFKQKTAYEQYF': #based on data preprocessing step in soNNia
        return False
    if filter_genes:
        if (v not in vs_default) or (j not in js_default):
            return False
    return True

if __name__ == '__main__':
    get_whole_data(['data_original/' + di for di in os.listdir('data_original')])



from utils import *
import pandas as pd
import numpy as np
from collections import defaultdict
from matplotlib.lines import Line2D
from evaluate import evaluation
from tcrpeg.TCRpeg import TCRpeg
import os
from scipy.stats import pearsonr as pr

path = '../../nlp_vae/data/whole_seqs_train_nn_sonnia.tsv'
# data = pd.read_csv(path, sep="\t",nrows=int(5e6))['seq'].values
#data = pd.read_csv(path, sep="\t").sample(n=int(200000),random_state=0)
#seqs,vs,js = data['seq'].values,data['v'].values,data['j'].values
model = TCRpeg(hidden_size=64,num_layers = 3,load_data=False,embedding_path='data/embedding_32.txt')
#model.aas_seqs_train= seqs
#model.vs_train,model.js_train = model.gene2embs(vs,'v'),model.gene2embs(js,'j')
model.create_model(vj=True,load=True,path='results/models/15:15:18/model.pth')
#model.train_tcrpeg_vj(30,32,5e-4,info='64_3_30_32_5e-4_200000_vj',record_dir='results/models')
for i in range(5):
    model.generate_tcrpeg_vj(200000, 2000,record_path='../../test_projects/results/generation_vj/tcrpeg_200000_{}.txt'.format(str(i)))



import torch
import numpy as np
import pandas as pd
import torch.nn as nn
from tcrpeg.model import TCRpeg_model, TCRpeg_vj_model
from tqdm import tqdm
import time
from datetime import datetime
import argparse
import os
from math import ceil

class TCRpeg:
    def __init__(
        self,
        max_length=30,
        embedding_size=32,
        hidden_size=64,
        dropout=0,
        num_layers=1,
        device="cuda:0",
        load_data=False,
        path_train=None,
        evaluate=0,
        path_test=None,
        embedding_path="data/embedding_32.txt",
        vs_list=None,
        js_list=None,
        vj=False
    ):
        '''
        @max_length: the maximum length of CDR3 seqs you want to deal with
        @embedding_size: length of the amino acid embedding vectors
        @hidden_size: The number of features in the hidden state of GRU layers
        @dropout: dropout rate
        @device: gpu or cpu you want the software run on; GPU: cuda:0' ('cuda:x') / CPU: 'cpu'
        @load_data & path_train: If you want to train tcrpeg, you should set load_data=True,path_train='path_to_training_file'
                                Also, it accepts list: [AAs1, AAs2,.....]
                                if using vj model, the format should be [[AAs seqs],[v genes],[j genes]]

        @evaluate & path_test : If you want to evaluate on test set, you should set evaluate=1, path_test='path_to_test_set'
                                Also accepts list as input. Same with 'path_train'
        @embedding_path: Recorded trained embeddings for amino acids (22 x embedding_size)
        @vs_list & js_list: The list of v and j genes. Only used when you're using the tcrpeg_vj model; 
                           The default v and j gene lists contain 43 and 12 genes respectively. If your TCR dataset
                           contains v,or j genes that are not included the default vs_list/js_list, yould should provide it.
                           type(vs_list) should be a list
        '''


        self.max_length = max_length
        self.embedding_size = embedding_size
        self.device = device
        self.dropout = dropout

        emb = np.array(
            pd.read_csv(embedding_path, sep=",", names=list(range(embedding_size + 1)))
        )
        embedding = np.zeros((23, embedding_size))  # 23 x letent_dim
        embedding[1:, :] = emb[:, 1:].astype(float)  # 22 x latent_dim
        self.embedding_layer = nn.Embedding.from_pretrained(
            torch.tensor(embedding), padding_idx=0
        ).to(device) #the embedding layer for AAs
        self.num_layers = num_layers

        self.aas = "sACDEFGHIKLMNPQRSTVWYe"  # 0 index is the padding,1 is START, 22 is END
        self.aa2idx = {self.aas[i]: i for i in range(len(self.aas))}
        self.idx2aa = {v: k for k, v in self.aa2idx.items()}
        
        vs_default = ['TRBV10-1','TRBV10-2','TRBV10-3','TRBV11-1','TRBV11-2','TRBV11-3','TRBV12-5', 'TRBV13', 'TRBV14', 'TRBV15', 
        'TRBV16', 'TRBV18','TRBV19','TRBV2', 'TRBV20-1', 'TRBV25-1', 'TRBV27', 'TRBV28', 'TRBV29-1', 'TRBV3-1', 'TRBV30',
'TRBV4-1', 'TRBV4-2','TRBV4-3',
 'TRBV5-1', 'TRBV5-4', 'TRBV5-5', 'TRBV5-6', 'TRBV5-8','TRBV6-1', 'TRBV6-4', 'TRBV6-5', 'TRBV6-6','TRBV6-8', 'TRBV6-9', 'TRBV7-2', 'TRBV7-3',
  'TRBV7-4', 'TRBV7-6', 'TRBV7-7', 'TRBV7-8', 'TRBV7-9', 'TRBV9']
        js_default = ['TRBJ1-1', 'TRBJ1-2', 'TRBJ1-3','TRBJ1-4', 'TRBJ1-5', 'TRBJ1-6','TRBJ2-1', 'TRBJ2-2', 'TRBJ2-3', 'TRBJ2-4', 'TRBJ2-6', 'TRBJ2-7']
        vs_default.sort() #to keep the order
        js_default.sort()

        if vs_list is not None:
            #here, to load the v and j genes you specify
            #the inputed vs and js needed to be sorted 
            vs_list.sort()
            js_list.sort()
            self.vs_list = vs_list
            self.js_list = js_list 
        else :
            self.vs_list = vs_default
            self.js_list = js_default
        self.v2idx = {self.vs_list[i]: i for i in range(len(self.vs_list))}
        self.idx2v = {v: k for k, v in self.v2idx.items()}
        self.j2idx = {self.js_list[i]: i for i in range(len(self.js_list))}
        self.idx2j = {v: k for k, v in self.j2idx.items()}
        self.loss_v,self.loss_j = nn.CrossEntropyLoss(reduction='sum'),nn.CrossEntropyLoss(reduction='sum')

        self.NLL = torch.nn.NLLLoss(
            ignore_index=-1, reduction="sum"
        ) #-1 index is the padding index in the training process, so need to ignore it  

        self.hidden_size = hidden_size

        self.evaluate = True if evaluate != 0 else False
        if load_data:
            if vj:
                if type(path_train) is not str:
                    self.aas_seqs_train = np.array([x[0] for x in path_train])
                    self.vs_train,self.js_train=[x[1] for x in path_train],[x[2] for x in path_train]
                else :
                    self.aas_seqs_train,self.vs_train,self.js_train = self.load_data(path_train,vj=True)
                self.vs_train,self.js_train = self.gene2embs(self.vs_train,'v'),self.gene2embs(self.js_train,'j')
            else :
                if type(path_train) is not str:
                    self.aas_seqs_train = path_train
                else :
                    self.aas_seqs_train = self.load_data(path_train)
            if self.evaluate:
                #remember to add!!
                if vj is not None:
                    if type(path_train) is not str:
                        self.aas_seqs_test = np.array([x[0] for x in path_test])
                        self.vs_test,self.js_test=[x[1] for x in path_test],[x[2] for x in path_test]
                    else :
                        self.aas_seqs_test,self.vs_test,self.js_test = self.load_data(path_test)
                    self.vs_test,self.js_test = self.gene2embs(self.vs_test,'v'),self.gene2embs(self.js_test,'j')
                else :
                    if type(path_test) is not str:
                        self.aas_seqs_test = path_test
                    else :
                        self.aas_seqs_test = self.load_data(path_test)
            print("Have loaded the data, total training seqs :", len(self.aas_seqs_train))

    def aas2embs(self, seqs):
        """
        @seqs: list of AAs

        #return [[1,2,3,0,0],[2,3,4,0].....], the padded 
        """
        seqs = [[self.aa2idx[k] + 1 for k in seq] for seq in seqs]
        # lengths = [len(seq)+2 for seq in seqs] #contains start and stop
        lengths = [len(seq) + 1 for seq in seqs]  # contains the start token
        inputs = [[1] + seq for seq in seqs]
        # inputs = [[1] + seq + [22] for seq in seqs]
        targets = [seq + [22] for seq in seqs]
        inputs = [seq + [0] * (self.max_length - len(seq)) for seq in inputs] #pad
        targets = [seq + [0] * (self.max_length - len(seq)) for seq in targets]#pad

        return inputs, targets, lengths  # targets are N x max_length

    def gene2embs(self, genes, gene_type='v'):
        '''
        @genes: a list containing genes
        '''
        if gene_type == 'v':
            return [self.v2idx[g] for g in genes]
        else :
            assert gene_type == 'j', 'the gene type should only be v or j'
            return [self.j2idx[g] for g in genes]

    def load_data(self, path,vj=False,sep='\t'):
        if vj:
            if type(path) == list:
                return path[0],path[1],path[2]
            
            data = pd.read_csv(path,sep)
            if data.columns[0] != 'seq':
                return data['amino_acid'].values,data['v_gene'].values,data['j_gene'].values    
            else :
                return data['seq'].values,data['v'].values,data['j'].values
        else :
            if type(path) == list:
                return path

            if not path.endswith("txt"):
                if 'vdj' in path: #this is used for loading epitope-tcrs
                    data = pd.read_csv(path)['cdr3b'].values
                    return data
                data = pd.read_csv(path, sep="\t")
            else:
                data = pd.read_csv(path, names=["seq"])
        return data["seq"].values

    def loss_fn(self, logp, target, lengths):
        '''
        Loss function for training a tcrpeg model
        '''
        logp = logp.view(-1, 22)
        target = target - 1
        # target is b x
        # lengths = lengths - 1 #remember to change this!
        target = target[:, : torch.max(lengths).item()].contiguous()

        target = target.view(-1)
        nll_loss = self.NLL(logp, target)

        return nll_loss

    def create_model(self, load=False, path=None,vj=False):
        '''
        Create the TCRpeg (TCRpeg_vj) model 
        @load: whether to load from pre-trained model
        @path: the path to pre-trained model, only need to specify when @load=True
        @vj: if set to True, will create the TCRpeg_vj model
        '''

        if vj:
            model = TCRpeg_vj_model(
                self.embedding_layer,
                self.embedding_size,
                self.hidden_size,
                self.dropout,
                self.max_length,
                num_layers=self.num_layers  ,
                num_v = len(self.vs_list),
                num_j = len(self.js_list)
            )
            self.vj=True
        else :
            model = TCRpeg_model(
                self.embedding_layer,
                self.embedding_size,
                self.hidden_size,
                dropout=self.dropout,
                num_layers=self.num_layers
                )
            self.vj=False
        model.train()
        if load:
            model.load_state_dict(torch.load(path))
            model.eval()
        self.model = model.to(self.device)
        # return model    

    def save(self, path):
        '''
        Save the model to the @path you specify
        '''
        torch.save(self.model.state_dict(), path)

    def sampling_tcrpeg(self, seqs):
        '''        
        @seqs: list containing CDR3 sequences
        #return: the log_prob of the input sequences
        '''
        with torch.no_grad():
            batch_size = len(seqs)

            inputs, targets, lengths = self.aas2embs(seqs)
            inputs, targets, lengths = (
                torch.LongTensor(inputs).to(self.device),
                torch.LongTensor(targets).to(self.device),
                torch.LongTensor(lengths).to(self.device),
            )
            logp = self.model(inputs, lengths)  
            
            logp = logp.detach().cpu()
            targets = targets.detach().cpu()

            targets = targets - 1

            target = targets[:, : torch.max(lengths).item()].contiguous()  # 0 is pad
            # logpx_z = torch.zeros(batch_size,torch.max(lengths).items())
            target = torch.where(target >= 0, target, 0)  # B x max_length ,0-21
            target_onehot = torch.FloatTensor(batch_size, torch.max(lengths).item(), 22)

            target_onehot.zero_()
            target_onehot.scatter_(
                2, target.view(batch_size, torch.max(lengths).item(), 1), 1
            )  # B x l x 22\

            probs = target_onehot * logp

            probs = probs.sum(dim=-1)  # B x L

            probs = torch.where(target > 0, probs, torch.FloatTensor([0.0])) #mask out the probs of padding positions
            logpx_given_z = probs.sum(dim=-1).numpy()
            return logpx_given_z  # B, log probability

    def sampling_tcrpeg_batch(self, seqs,batch_size=10000):
        '''
        @seqs: list containing CDR3 sequences
        #return: the log_prob of the input sequences
        Inferring in batch
        '''
        logpx_given_z = np.zeros(len(seqs))
        with torch.no_grad():
            #batch_size = len(seqs)
            for i in tqdm(range(int(len(seqs)/batch_size)+1)):
                end = len(seqs) if (i+1) * batch_size > len(seqs) else (i+1) * batch_size
                seq_batch = seqs[i * batch_size : end]
                if len(seq_batch) == 0:
                    continue
                log_probs = self.sampling_tcrpeg(seq_batch) #change here
                logpx_given_z[i*batch_size : end] = log_probs
            return logpx_given_z  # B, log probability



    def train_tcrpeg(self, epochs, batch_size, lr, info=None, model_name=None,record_dir=None):
        '''
        @epochs: epochs
        @batch_size: batch_size
        @lr: initial learning rate; The learning rate will reduced by lr=lr*0.2 at the middle of training
        @info: the information you want to record at the top of the log file (only activated when you specify the record dir)
        @model_name: the models will be saved as model_name.pth (only activated when you specify the record_dir)
        @record_dir: the directory you want to record your models; if not provided, the trained models will not be saved
        '''
        print("begin the training process")
        optimizer = torch.optim.Adam(self.model.parameters(), lr=lr)
        batch_size = batch_size
        epochs = epochs
        record = True if record_dir is not None else False
        if info is None:
            info = 'Not provided'
        if model_name is None:
            model_name = 'Not provided'
        

        # start_time = datetime.now().strftime("%m_%d_%Y-%H:%M:%S")
        if record:
            # dir = record_dir + "/" + start_time.split("-")[1]
            dir = record_dir 
            # if not os.path.exists(dir):
            #     os.mkdir(dir)
            with open(dir + "/logs.txt", "a") as f:
                f.write(info + "\n")
        aas_seqs = self.aas_seqs_train

        
        for epoch in range(epochs):
            print("begin epoch :", epoch + 1)
            self.model.train()
            aas_seqs = np.random.permutation(aas_seqs) #shuffle at the beginning of each epoch
            nll_loss_ = []
            num_batch = len(aas_seqs) // batch_size
            for iter in tqdm(range(len(aas_seqs) // batch_size)):
                # seqs = aas_seqs[infer_arr[iter * batch_size : (iter+1) * batch_size]]
                seqs = aas_seqs[iter * batch_size : (iter + 1) * batch_size]
                inputs, targets, lengths = self.aas2embs(seqs) #embeds the sequences
                inputs, targets, lengths = (
                    torch.LongTensor(inputs).to(self.device),
                    torch.LongTensor(targets).to(self.device),
                    torch.LongTensor(lengths).to(self.device),
                )
                logp = self.model(inputs, lengths)

                nllloss = self.loss_fn(logp, targets, lengths)
                nll_loss_.append(nllloss.item() / batch_size)
                loss = nllloss / batch_size  # remember to multiply the beta!
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
            print("nll_loss: {0}".format(sum(nll_loss_) / num_batch))
            if record:
                with open(dir + "/logs.txt", "a") as f:
                    f.write(
                        "For {} model, trained at {}_th epoch with nll (train) are {}\n".format(
                            model_name,
                            epoch + 1,
                            sum(nll_loss_) / num_batch,
                        )
                    )
        
                if epoch % 3 == 0:
                    # early_stopping.save_checkpoint(sum(elbo_test)/num_batch,model)
                    print('wrong enter')
                    self.save(
                        dir
                        + "/"
                        + model_name
                        + "_{}.pth".format(str(epoch + 1))
                    )
    
            if epoch != 0 and epoch % ((epochs) // 2) == 0 :
                print('The learning rate has beed reduced')
                optimizer.param_groups[0]["lr"] = lr * 0.2

        print("Done training")
        if record:
            self.save(
                dir + "/" + model_name + "_{}.pth".format(str(epoch + 1))
            )
    
    def evaluate(self,batch_size,vj=False):
        '''
        Evaluation on test set for tcrpeg model
        @batch_size: batch_size for loading data
        @vj: whether the model is TCRpeg or TCRpeg_vj

        @return: (nll_loss,v_loss,j_loss) if vj is True else return the nll_loss
        '''
        aas_val = self.aas_seqs_test
        with torch.no_grad():
            nll_test = []
            v_test_,j_test_ = [],[]
            num_batch = len(aas_val) // batch_size
            for iter in tqdm(range(num_batch)):
                seqs = aas_val[iter * batch_size : (iter + 1) * batch_size]
                if vj:
                    vs,js = self.vs_test[iter * batch_size : (iter + 1) * batch_size],self.js_test[iter * batch_size : (iter + 1) * batch_size]
                    vs,js = torch.LongTensor(vs).to(self.device),torch.LongTensor(js).to(self.device)
                inputs, targets, lengths = self.aas2embs(seqs)
                inputs, targets, lengths = (
                    torch.LongTensor(inputs).to(self.device),
                    torch.LongTensor(targets).to(self.device),
                    torch.LongTensor(lengths).to(self.device),
                )
                if vj:
                    logp,v_pre,j_pre = self.model(inputs,lengths)
                else :
                    logp = self.model(inputs, lengths)

                nllloss = self.loss_fn(
                    logp, targets, lengths, decoder_only=True
                )
                if vj:
                    v_loss,j_loss = self.loss_v(v_pre,vs),self.loss_j(j_pre,js)
                    v_test_.append(v_loss.item()/batch_size)
                    j_test_.append(j_loss.item()/batch_size)
                nll_test.append(nllloss.item() / batch_size)
        if vj:
            print("nll_loss,v_loss,j_loss for test set: {0}, {1}, {2}".format(sum(nll_test) / num_batch,sum(v_test_) / num_batch,sum(j_test_) / num_batch))
            return sum(nll_test) / num_batch,sum(v_test_) / num_batch,sum(j_test_) / num_batch

        else :
            print("nll_loss for val: {0}".format(sum(nll_test) / num_batch))
            return sum(nll_test) / num_batch
        

    def train_tcrpeg_vj(self, epochs, batch_size, lr, info=None, model_name=None,record_dir=None):
        '''
        Train a TCRpeg with v j usage
        Same parameter setting as @train_tcrpeg
        '''
        print("begin the training process")
        optimizer = torch.optim.Adam(self.model.parameters(), lr=lr)
        batch_size = batch_size
        epochs = epochs
        aas_seqs = self.aas_seqs_train
        vs_,js_ = np.array(self.vs_train),np.array(self.js_train)
        record = True if record_dir is not None else False
        #start_time = datetime.now().strftime("%m_%d_%Y-%H:%M:%S")
        if info is None:
            info = 'Not provided'
        if model_name is None:
            model_name = 'Not provided'

        if record:
            #dir = record_dir + "/" + start_time.split("-")[1]
            dir = record_dir
            # if not os.path.exists(dir):
            #     os.mkdir(dir)
            with open(dir + "/logs.txt", "a") as f:
                f.write(info + "\n")
        infer = np.arange(len(self.aas_seqs_train)) #used to shuffle the training data

        
    
        for epoch in range(epochs):
            np.random.shuffle(infer)
            print("begin epoch :", epoch + 1)
            self.model.train()
            
            aas_seqs = aas_seqs[infer]
            vs_,js_ = vs_[infer],js_[infer] #shuffle
            nll_loss_ = []
            j_loss_ = []
            v_loss_ = []
            num_batch = len(aas_seqs) // batch_size
            for iter in tqdm(range(len(aas_seqs) // batch_size)):
                # seqs = aas_seqs[infer_arr[iter * batch_size : (iter+1) * batch_size]]
                seqs = aas_seqs[iter * batch_size : (iter + 1) * batch_size]
                vs = vs_[iter * batch_size : (iter + 1) * batch_size]
                js = js_[iter * batch_size : (iter + 1) * batch_size]
                vs ,js = torch.LongTensor(vs).to(self.device),torch.LongTensor(js).to(self.device)

                inputs, targets, lengths = self.aas2embs(seqs)
                inputs, targets, lengths = (
                    torch.LongTensor(inputs).to(self.device),
                    torch.LongTensor(targets).to(self.device),
                    torch.LongTensor(lengths).to(self.device),
                )
                logp,v_pre,j_pre = self.model(inputs, lengths)

                nllloss = self.loss_fn(logp, targets, lengths)
                v_loss,j_loss = self.loss_v(v_pre,vs),self.loss_j(j_pre,js)
                nll_loss_.append(nllloss.item() / batch_size)
                v_loss_.append(v_loss.item()/batch_size)
                j_loss_.append(j_loss.item()/batch_size)
                loss = nllloss / batch_size  + 1.5 * v_loss/batch_size + 1.5 * j_loss/batch_size # remember to multiply the beta!
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
            print("nll_loss: {0}".format(sum(nll_loss_) / num_batch))
            print("v_loss: {0}".format(sum(v_loss_) / num_batch))
            print("j_loss: {0}".format(sum(j_loss_) / num_batch))
            if record:
                with open(dir + "/logs.txt", "a") as f:
                    f.write(
                        "For {} model, trained at {}_th epoch with nll (train),v,j are {},{},{}\n".format(
                            model_name,
                            epoch + 1,
                            sum(nll_loss_) / num_batch,
                            sum(v_loss_) / num_batch,
                            sum(j_loss_) / num_batch
                        )
                    )
                if epoch % 3 == 0:
                    # early_stopping.save_checkpoint(sum(elbo_test)/num_batch,model)
                    self.save(
                        dir
                        + "/"
                        + model_name
                        + "_{}.pth".format(str(epoch + 1))
                    )
            if epoch != 0 and epoch % ((epochs) // 2) == 0:
                print('The learning rate has beed reduced')
                optimizer.param_groups[0]["lr"] = lr * 0.2

        print("Done training")
        if record:
            self.save(
                dir + "/" + model_name + "_{}.pth".format( str(epoch + 1))
            )

    def generate_decoder(self, num_to_gen, record_path=None):
        # need to change this to batch!!!  20k seqs need ~2 mins
        SOS_token = 1
        EOS_token = 22
        self_ = self
        self = self.model
        refer = list(range(22))  # 0~21
        # self = self.model
        seqs = []
        batch_size = 1
        with torch.no_grad():
            for _ in tqdm(range(num_to_gen)):
                decoder_input = torch.LongTensor([[SOS_token] for _ in range(batch_size)]).to("cuda:0")
                seq = []
                for t in range(30):
                    decoder_input = (
                        self.embedding(decoder_input).to(torch.float32).permute([1, 0, 2]))
                    if t != 0:
                        decoder_output, decoder_hidden = self.decoder_rnn(decoder_input, decoder_hidden)
                    else:
                        decoder_output, decoder_hidden = self.decoder_rnn(decoder_input)
                    decoder_output = self.out_layer(decoder_output)
                    decoder_output = nn.functional.softmax(decoder_output, dim=-1).view(
                        -1
                    )  # 22,
                    next_idx = np.random.choice(refer, p=decoder_output.detach().cpu().numpy())
                    if next_idx == EOS_token - 1:  ##
                        break
                    seq.append(self_.idx2aa[next_idx])
                    decoder_input = torch.LongTensor([[next_idx + 1] for _ in range(batch_size)]).to("cuda:0")
                seqs.append("".join(seq))

        if record_path is not None:
            with open(record_path, "w") as f:
                for s in seqs:
                    f.write(s + "\n")

    def generate_tcrpeg(self, num_to_gen, batch_size,record_path=None):
        '''
        Generating new sequences by the trained model; Note that (num_to_gen % batch_size) should be 0
        @num_to_gen: number of sequences need to generate
        @batch_size: batch_size
        @record_path: if specified, will record the generated seqs 
        #return: a list containing the generated seqs
        '''
        # need to change this to batch!!!  20k seqs need ~2 mins
        SOS_token = 1
        EOS_token = 22
        self_ = self
        self = self.model
        
        # self = self.model
        seqs = []
        assert num_to_gen % batch_size == 0, 'The num_to_gen have to be a multiple of batch_size'
        steps = num_to_gen // batch_size
        with torch.no_grad():
            for _ in tqdm(range(steps)):
                decoded_batch = torch.zeros((batch_size, self.max_length))
                decoder_input = torch.LongTensor([[SOS_token] for _ in range(batch_size)]).to(self.device)
                EOS_position = np.ones(batch_size,dtype=int) * -1
                
                for t in range(self.max_length):
                    decoder_input = self.embedding(decoder_input).to(torch.float32).permute([1,0,2])
                    if t != 0:
                        decoder_output, decoder_hidden = self.decoder_rnn(decoder_input, decoder_hidden)
                    else:
                        decoder_output, decoder_hidden = self.decoder_rnn(decoder_input)
                    decoder_output = self.out_layer(decoder_output)
                    decoder_output = nn.functional.softmax(decoder_output, dim=-1).view(batch_size,22)  # 1 xB x 22,
                    #print(decoder_output.size())
                    next_idx = torch.multinomial(decoder_output,1,True).view(-1) #B,
                    for k in range(batch_size):
                        if next_idx[k] == EOS_token - 1 and EOS_position[k] == -1:
                            EOS_position[k] = t
                    # if next_idx == EOS_token - 1:  ##
                    #     break
                    decoded_batch[:,t] = next_idx
                    decoder_input = (next_idx + 1).view(-1,1)
                decoded_batch = decoded_batch.cpu().numpy().astype(int)
                decoded = []
                for k in range(batch_size):
                #print(EOS_position[k])
                    decoded.append(list(decoded_batch[k,:EOS_position[k]]))
                decoded = [[self_.idx2aa[a] for a in seq] for seq in decoded]
                decoded = [''.join(seq) for seq in decoded]
                seqs = seqs + decoded
            
        if record_path is not None:
            with open(record_path, "w") as f:
                for s in seqs:
                    f.write(s + "\n")
        return seqs

    def generate_tcrpeg_vj(self, num_to_gen, batch_size,record_path=None):
        '''
        #return: a list containing the generated seqs,vs,js in the format of [[seqs],[vs],[js]]
        '''
        # need to change this to batch!!!  20k seqs need ~2 mins
        SOS_token = 1
        EOS_token = 22
        self_ = self
        self = self.model
        
        # self = self.model
        seqs = []
        assert num_to_gen % batch_size == 0, 'The num_to_gen have to be a multiple of batch_size'
        steps = num_to_gen // batch_size
        with torch.no_grad():
            for _ in tqdm(range(steps)):
                decoded_batch = torch.zeros((batch_size, self.max_length))
                decoder_input = torch.LongTensor([[SOS_token] for _ in range(batch_size)]).to(self.device)
                EOS_position = np.ones(batch_size,dtype=int) * -1
                
                for t in range(self.max_length):
                    decoder_input = self.embedding(decoder_input).to(torch.float32).permute([1,0,2])
                    if t != 0:
                        decoder_output, decoder_hidden = self.decoder_rnn(decoder_input, decoder_hidden)
                    else:
                        decoder_output, decoder_hidden = self.decoder_rnn(decoder_input)
                    decoder_output = self.out_layer(decoder_output)
                    decoder_output = nn.functional.softmax(decoder_output, dim=-1).view(batch_size,22)  # 1 xB x 22,
                    #print(decoder_output.size())
                    next_idx = torch.multinomial(decoder_output,1,True).view(-1) #B,
                    for k in range(batch_size):
                        if next_idx[k] == EOS_token - 1 and EOS_position[k] == -1:
                            EOS_position[k] = t
                    # if next_idx == EOS_token - 1:  ##
                    #     break
                    decoded_batch[:,t] = next_idx
                    decoder_input = (next_idx + 1).view(-1,1)
                decoded_batch = decoded_batch.cpu().numpy().astype(int)
                decoded = []
                for k in range(batch_size):
                #print(EOS_position[k])
                    decoded.append(list(decoded_batch[k,:EOS_position[k]]))
                decoded = [[self_.idx2aa[a] for a in seq] for seq in decoded]
                decoded = [''.join(seq) for seq in decoded]
                seqs = seqs + decoded
            
            #used the generated seqs to generate the v and j gene:
            vs_whole,js_whole = [],[]
            for iter in tqdm(range(steps)):
                seqs_batch = seqs[iter * batch_size : (iter + 1) * batch_size]
                inputs, targets, lengths = self_.aas2embs(seqs_batch)
                inputs, targets, lengths = (
                    torch.LongTensor(inputs).to(self_.device),
                    torch.LongTensor(targets).to(self_.device),
                    torch.LongTensor(lengths).to(self_.device),
                )
                _,v_pre,j_pre = self_.model(inputs, lengths)
                v_pre,j_pre = nn.functional.softmax(v_pre,-1),nn.functional.softmax(j_pre,-1)
                v_pre_id = torch.multinomial(v_pre,1,True).view(-1).detach().cpu().numpy() #B,
                j_pre_id = torch.multinomial(j_pre,1,True).view(-1).detach().cpu().numpy()

                v_pres = [self_.idx2v[i] for i in v_pre_id]
                j_pres = [self_.idx2j[i] for i in j_pre_id]
                vs_whole = vs_whole + list(v_pres)
                js_whole = js_whole + list(j_pres)
            
        if record_path is not None:
            with open(record_path, "w") as f:
                for i,s in enumerate(seqs):
                    f.write(s + ',' + vs_whole[i] + ',' + js_whole[i] + "\n")
        return [seqs,vs_whole,js_whole]

    def get_embedding(self,seqs,last_layer=False):
        '''
        Get the embedding of CDR3 sequences
        @seqs: a list containing the CDR3 sequences
        @batch_size: batch_size
        @last_layer: if set to True, will return the hidden features of the last GRU layer. Otherwise, hidden features of all layers are used

        #return: embedding of CDR3 sequences. The shape would be (B,num_layers*hidden_size) if last_layer=False, (B,hidden_size) if last_layer=True
        '''
        nums = len(seqs)
        self.model.eval()
        with torch.no_grad():
            inputs,targets,lengths = self.aas2embs(seqs)
            inputs,targets,lengths = torch.LongTensor(inputs).to(self.device),torch.LongTensor(targets).to(self.device),torch.LongTensor(lengths).to(self.device)
            if self.vj:
                _,_,_,embedding = self.model(inputs,lengths,True)
            else :
                _,embedding= self.model(inputs,lengths,True)
            if last_layer:
                embedding = embedding[:,-1,:] #B x 64
            else :
                embedding = embedding.view(nums,-1)
            #embedding = embedding.view(batch_size,-1)
        return embedding.detach().cpu().numpy()




from tcrpeg.TCRpeg import TCRpeg
from tcrpeg.classification import classification
import pandas as pd
import numpy as np
import os
from tcrpeg.evaluate import evaluation
from tcrpeg.utils import plotting
from tcrpeg.word2vec import word2vec
from sklearn.model_selection import KFold as kfold

vs_default = ['TRBV10-1','TRBV10-2','TRBV10-3','TRBV11-1','TRBV11-2','TRBV11-3','TRBV12-3','TRBV12-5', 'TRBV13', 'TRBV14', 'TRBV15', 
        'TRBV16', 'TRBV18','TRBV19','TRBV2', 'TRBV20-1', 'TRBV25-1', 'TRBV27', 'TRBV28', 'TRBV29-1', 'TRBV3-1', 'TRBV30',
'TRBV4-1', 'TRBV4-2','TRBV4-3','TRBV6-3',
 'TRBV5-1', 'TRBV5-4', 'TRBV5-5', 'TRBV5-6', 'TRBV5-8','TRBV6-1', 'TRBV6-2','TRBV6-4', 'TRBV6-5','TRBV12-4', 'TRBV6-6','TRBV6-8','TRBV24-1', 'TRBV6-9', 'TRBV7-2', 'TRBV7-3',
  'TRBV7-4', 'TRBV7-6', 'TRBV7-7', 'TRBV7-8', 'TRBV7-9', 'TRBV9']
js_default = ['TRBJ1-1', 'TRBJ1-2', 'TRBJ1-3','TRBJ1-4', 'TRBJ1-5', 'TRBJ1-6','TRBJ2-1', 'TRBJ2-2', 'TRBJ2-3', 'TRBJ2-4', 'TRBJ2-5','TRBJ2-6', 'TRBJ2-7']
v,j = [x[4:] for x in vs_default],[x[4:] for x in js_default]

#get embeddings:
#np.random.seed(0)
# pos = list(pd.read_csv('data/classification/GILGFVFTL/positive.txt',names=['seq'])['seq'].values)
# con = list(pd.read_csv('data/classification/GILGFVFTL/control.txt',names=['seq'])['seq'].values)
whole = (pd.read_csv('data/classification/GLCTLVAML/N_whole.csv')[['CDR3_sequence','V_gene','J_gene']]).values
# whole = np.array(pos + con)
whole_y = np.array([1]* (len(whole)//2) + [0] * (len(whole)//2))
# model = TCRpeg(hidden_size=256,num_layers = 3,load_data=True,embedding_path='data/embedding_32.txt',path_train=whole)
# model.create_model()
# model.train_tcrpeg(epochs=30,batch_size=8,lr=1e-3)
# model.save('results/additional/G.pth')

# pos = list(pd.read_csv('data/classification/NLVPMVATV/positive.txt',names=['seq'])['seq'].values)
# con = list(pd.read_csv('data/classification/NLVPMVATV/control.txt',names=['seq'])['seq'].values)
# whole = pos + con
# model = TCRpeg(hidden_size=256,num_layers = 3,load_data=True,embedding_path='data/embedding_32.txt',path_train=whole)
# model.create_model()
# model.train_tcrpeg(epochs=30,batch_size=8,lr=1e-3)
# model.save('results/additional/N.pth')

kf = kfold(n_splits=5,random_state=0,shuffle=True)
index = 0
for train_idx,test_idx in kf.split(whole):
    test,test_y = whole[test_idx],whole_y[test_idx]
    test = np.array([x[0] for x in test])
    train,train_y = whole[train_idx],whole_y[train_idx]
    # t_model = TCRpeg(hidden_size=256,num_layers = 3,load_data=True,embedding_path='data/embedding_32.txt',path_train=train,vs_list=v,js_list=j,vj=True)
    # t_model.create_model(vj=True)
    # t_model.train_tcrpeg_vj(epochs=30,batch_size=8,lr=1e-3)
    # t_model.save('results/additional/models/N_vj_{}.pth'.format(index))
    #####
    train_pos = [train[i] for i in range(len(train)) if train_y[i] == 1]
    train_neg = [train[i] for i in range(len(train)) if train_y[i] == 0]
    model_pos = TCRpeg(hidden_size=256,num_layers = 3,load_data=True,embedding_path='data/embedding_32.txt',path_train=train_pos,vs_list=v,js_list=j,vj=True)
    model_pos.create_model(vj=True)
    model_pos.train_tcrpeg_vj(epochs=30,batch_size=8,lr=1e-3)
    model_pos.save('results/additional/models/N_pos_vj_{}.pth'.format(index))

    model_neg = TCRpeg(hidden_size=256,num_layers = 3,load_data=True,embedding_path='data/embedding_32.txt',path_train=train_neg,vs_list=v,js_list=j,vj=True)
    model_neg.create_model(vj=True)
    model_neg.train_tcrpeg_vj(epochs=30,batch_size=8,lr=1e-3)
    model_neg.save('results/additional/models/N_neg_vj_{}.pth'.format(index))

    # tcrpeg_c = classification(t_model,256*3,dropout=0.2,model_size='large')
    # tcrpeg_c.train([x[0] for x in train],train_y,25,8,1e-3,val_split=0,save_name=None)
    # auc,aup,y_pres,y_trues = tcrpeg_c.evaluate(test,test_y,100,record_path='results/additional/G_vj_{}.txt'.format(index))
    # with open('results/additional/predictions/pres_large_vj_{}.txt'.format(str(index)),'w') as f:
    #     for i in range(len(y_pres)):
    #         f.write(str(y_pres[i])+','+str(y_trues[i]) + '\n')
    # test_pos = [test[i] for i in range(len(test)) if test_y[i] == 1]
    # test_neg = [test[i] for i in range(len(test)) if test_y[i] == 0]
    # with open('data/classification/GILGFVFTL/deepcat/G_normal_train_{}.txt')

    index += 1


from utils import *
import pandas as pd
import numpy as np
from collections import defaultdict
from matplotlib.lines import Line2D
from evaluate import evaluation
from tcrpeg.TCRpeg import TCRpeg
import os
from scipy.stats import pearsonr as pr

Plot = plotting()
# model = TCRpeg(hidden_size=64,num_layers = 3,load_data=False,embedding_path='data/embedding_32.txt')
# model.create_model(load=True,path='../../nlp_vae/results/models/11_22_2021/19:35:40/only_decoder_11_22_2021-19:35:40_40.pth')
# tcrpeg_path = '../../nlp_vae/results/models/12_20_2021/17:00:47/none_12_20_2021-17:00:47_20.pth'
# t_model = TCRpeg(hidden_size=256,num_layers = 3,load_data=False,embedding_path='data/embedding_32.txt',vj=True)
# t_model.create_model(load=True,vj=True,path=tcrpeg_path)

#eva = evaluation(model=model,frac=0.1)
def compute_ld(seqs):
    '''
    
    '''
    l1 = len(seqs)
    seqs = [s for s in seqs if Plot.valid_seq(s)]
    lens = [len(seq) for seq in seqs]
    len2count = defaultdict(int)
    print('diff:', l1 - len(seqs))
    for i in range(len(lens)):    
        len2count[lens[i]] += 1
    for key in len2count.keys():
        len2count[key] /= len(lens)
    return len2count

def compute_laa(seqs):
    aa2pos_dic = [defaultdict(int) for _ in range(20)]
    seqs = [s for s in seqs if Plot.valid_seq(s)]
    for i in range(len(seqs)):
        seq1 = seqs[i]
        for j,aa in enumerate(seq1):
            #if aa == 'e' or 
            aa2pos_dic[Plot.aa2idx[aa]][j+1] += 1 #remember to modify
    x_axis = list(range(1,31))
    for i in range(20):
        dic = aa2pos_dic[i]
        y = [dic[k] for k in x_axis]
        sum_ = sum(y)
        y_nor = [yy/sum_ for yy in y]
        #dic[i] = [yy/sum_ for yy in y]
        aa2pos_dic[i] = {x_axis[i]:y_nor[i] for i in range(len(x_axis))}
        # print(dic[i])
        # print(aa2pos_dic[i])
    
    return aa2pos_dic

def compute_v(vs):
    vs_dic = defaultdict(int)
    for i in range(len(vs)):
        vs_dic[vs[i]] += 1
    x_axis = Plot.vs
    y1 = [vs_dic[x]/len(vs) for x in x_axis]
    return {x_axis[i]:y1[i] for i in range(len(x_axis))}

def compute_j(js):
    vs_dic = defaultdict(int)
    for i in range(len(js)):
        vs_dic[js[i]] += 1
    x_axis = Plot.js
    
    y1 = [vs_dic[x]/len(js) for x in x_axis]
    
    
    return {x_axis[i]:y1[i] for i in range(len(x_axis))}

d = pd.read_csv('../../nlp_vae/data/whole_seqs_test_nn_sonnia.tsv',sep='\t')
#len2count_total = defaultdict(list)

ss,vs,js = d['seq'].values,d['v'].values,d['j'].values
v_dic_data = compute_v(vs)
j_dic_data = compute_j(js)

#record kl
kl_peg_aa,kl_sonnia_aa,kl_tcrvae_aa = [[]]*20,[[]]*20,[[]]*20
kl_peg_l,kl_sonnia_l,kl_tcrvae_l = [],[],[]
kl_peg_v,kl_sonnia_v,kl_tcrvae_v = [],[],[]
kl_peg_j,kl_sonnia_j,kl_tcrvae_j = [],[],[]

##length dis
total_lendis = dict()
total_aasdis = dict()
total_vsdis = dict()
total_jsdis = dict()

dirs_base = '../../test_projects/results/generation_vj/'
dirs = [dirs_base + x for x in os.listdir(dirs_base) if 'peg' in x and '200000' in x]
assert len(dirs) == 5
len2count_total = defaultdict(list)
aa2count_total = [defaultdict(list) for _ in range(20)]
v_total = defaultdict(list)
j_total = defaultdict(list)
for i in range(5):
    #data_path = '../../nlp_vae/results/gens/gens_200000/vae_200000.csv'
    data_path = dirs[i]
    print(data_path)
    d = pd.read_csv(data_path,names=['seqs','v','j'])
    ss,vs,js = d['seqs'].values,d['v'].values,d['j'].values
    len2count = compute_ld(ss)
    aa2pos_dic = compute_laa(ss)
    j_dic = compute_j(js)
    for key in j_dic.keys():
        j_total[key].append(j_dic[key])
    for a in range(20):
        aa2pos = aa2pos_dic[a]
        for key in aa2pos.keys():
            aa2count_total[a][key].append(aa2pos[key])
    v_dic = compute_v(vs)
    for key in v_dic.keys():
        v_total[key].append(v_dic[key])
    for l in range(2,31):
        len2count_total[l].append(len2count[l]) #to change
for i in range(20):
    aa2count = aa2count_total[i]
    for key in aa2count.keys():
        array = aa2count[key]
        aa2count[key] = (np.mean(array),np.std(array))

v_list = list(v_total.keys())
ks = []
for i in range(5):
    y1 = [v_dic_data[v] for v in v_list]
    y2 = [v_total[v][i] for v in v_list]
    # ks.append(kl_divergence(y1,y2,True))
    ks.append(pr(y1,y2)[0])
print(np.mean(ks))
print(np.std(ks))

j_list = list(j_total.keys())
ks = []
for i in range(5):
    y1 = [j_dic_data[v] for v in j_list]
    y2 = [j_total[v][i] for v in j_list]
    # ks.append(kl_divergence(y1,y2,True))
    ks.append(pr(y1,y2)[0])
print(np.mean(ks))
print(np.std(ks))

v_dis_input = ([np.mean(v_total[x]) for x in Plot.vs],[np.std(v_total[x]) for x in Plot.vs])
j_dis_input = ([np.mean(j_total[x]) for x in Plot.js],[np.std(j_total[x]) for x in Plot.js])
total_lendis['tcrpeg'] = len2count_total
total_aasdis['tcrpeg'] = aa2count_total
total_vsdis['tcrpeg'] = v_dis_input
total_jsdis['tcrpeg']= j_dis_input
###
dirs_base = '../../test_projects/results/generation_vj/'
dirs = [dirs_base + x for x in os.listdir(dirs_base) if 'sonnia' in x and '20000' in x]
assert len(dirs) == 5
len2count_total = defaultdict(list)
aa2count_total = [defaultdict(list) for _ in range(20)]
v_total = defaultdict(list)
j_total = defaultdict(list)
for i in range(5):
    #data_path = '../../nlp_vae/results/gens/gens_200000/vae_200000.csv'
    data_path = dirs[i]
    print(data_path)
    d = pd.read_csv(data_path,names=['seqs','v','j'])
    ss,vs,js = d['seqs'].values,d['v'].values,d['j'].values
    len2count = compute_ld(ss)
    aa2pos_dic = compute_laa(ss)
    v_dic = compute_v(vs)
    j_dic = compute_j(js)
    for key in j_dic.keys():
        j_total[key].append(j_dic[key])
    for key in v_dic.keys():
        v_total[key].append(v_dic[key])
    for a in range(20):
        aa2pos = aa2pos_dic[a]
        for key in aa2pos.keys():
            aa2count_total[a][key].append(aa2pos[key])
    for l in range(2,31):
        len2count_total[l].append(len2count[l]) #to change
for i in range(20):
    aa2count = aa2count_total[i]
    for key in aa2count.keys():
        array = aa2count[key]
        aa2count[key] = (np.mean(array),np.std(array))

v_list = list(v_total.keys())
ks = []
for i in range(5):
    y1 = [v_dic_data[v] for v in v_list]
    y2 = [v_total[v][i] for v in v_list]
    #ks.append(kl_divergence(y1,y2,True))
    ks.append(pr(y1,y2)[0])
print(np.mean(ks))
print(np.std(ks))

j_list = list(j_total.keys())
ks = []
for i in range(5):
    y1 = [j_dic_data[v] for v in j_list]
    y2 = [j_total[v][i] for v in j_list]
    #ks.append(kl_divergence(y1,y2,True))
    ks.append(pr(y1,y2)[0])
print(np.mean(ks))
print(np.std(ks))

v_dis_input = ([np.mean(v_total[x]) for x in Plot.vs],[np.std(v_total[x]) for x in Plot.vs])
j_dis_input = ([np.mean(j_total[x]) for x in Plot.js],[np.std(j_total[x]) for x in Plot.js])
total_lendis['sonnia'] = len2count_total
total_aasdis['sonnia'] = aa2count_total
total_vsdis['sonnia'] = v_dis_input
total_jsdis['sonnia'] = j_dis_input
###
dirs_base = '../../test_projects/results/generation_vj/'
dirs = [dirs_base + x for x in os.listdir(dirs_base) if 'vae' in x and '20k' in x]
assert len(dirs) == 5
len2count_total = defaultdict(list)
aa2count_total = [defaultdict(list) for _ in range(20)]
v_total = defaultdict(list)
j_total = defaultdict(list)
for i in range(5):
    #data_path = '../../nlp_vae/results/gens/gens_200000/vae_200000.csv'
    data_path = dirs[i]
    print(data_path)
    d = pd.read_csv(data_path)
    ss,vs,js = d['seqs'].values,d['v'].values,d['j'].values
    len2count = compute_ld(ss)
    aa2pos_dic = compute_laa(ss)
    v_dic = compute_v(vs)
    j_dic = compute_j(js)
    for key in j_dic.keys():
        j_total[key].append(j_dic[key])
    for key in v_dic.keys():
        v_total[key].append(v_dic[key])
    for a in range(20):
        aa2pos = aa2pos_dic[a]
        for key in aa2pos.keys():
            aa2count_total[a][key].append(aa2pos[key])
    for l in range(2,31):
        len2count_total[l].append(len2count[l]) #to change

v_list = list(v_total.keys())
ks = []
for i in range(5):
    y1 = [v_dic_data[v] for v in v_list]
    y2 = [v_total[v][i] for v in v_list]
    #ks.append(kl_divergence(y1,y2,True))
    ks.append(pr(y1,y2)[0])
print(np.mean(ks))
print(np.std(ks))

j_list = list(j_total.keys())
ks = []
for i in range(5):
    y1 = [j_dic_data[v] for v in j_list]
    y2 = [j_total[v][i] for v in j_list]
    #ks.append(kl_divergence(y1,y2,True))
    ks.append(pr(y1,y2)[0])
print(np.mean(ks))
print(np.std(ks))

v_dis_input = ([np.mean(v_total[x]) for x in Plot.vs],[np.std(v_total[x]) for x in Plot.vs])
j_dis_input = ([np.mean(j_total[x]) for x in Plot.js],[np.std(j_total[x]) for x in Plot.js])
for i in range(20):
    aa2count = aa2count_total[i]
    for key in aa2count.keys():
        array = aa2count[key]
        aa2count[key] = (np.mean(array),np.std(array))
total_lendis['tcrvae'] = len2count_total
total_aasdis['tcrvae'] = aa2count_total
total_vsdis['tcrvae'] = v_dis_input
total_jsdis['tcrvae'] = j_dis_input
###
d = pd.read_csv('../../nlp_vae/data/whole_seqs_test_nn_sonnia.tsv',sep='\t')
#len2count_total = defaultdict(list)

ss,vs,js = d['seq'].values,d['v'].values,d['j'].values
aa2pos_dic = compute_laa(ss)
len2count = compute_ld(ss)
v_dic = compute_v(vs)
j_dic = compute_j(js)
vs_ = [v_dic[x] for x in Plot.vs]
js_ = [j_dic[x] for x in Plot.js]
data_v = (vs_,vs_)
data_j = (js_,js_)
data = [len2count[l] for l in range(2,31)]

# #length
seqs1 = np.array([np.mean(total_lendis['tcrpeg'][i]) for i in range(2,31)])
errs1 = np.array([(np.std(total_lendis['tcrpeg'][i]) ) for i in range(2,31)])
seqs2 = np.array([np.mean(total_lendis['sonnia'][i]) for i in range(2,31)])
errs2 = np.array([(np.std(total_lendis['sonnia'][i]) ) for i in range(2,31)])
seqs3 = np.array([np.mean(total_lendis['tcrvae'][i]) for i in range(2,31)])
errs3 = np.array([(np.std(total_lendis['tcrvae'][i]) ) for i in range(2,31)])
Plot.length_dis(data,(seqs1,errs1),(seqs2,errs2),(seqs3,errs3),list(range(2,31)),fig_name='length_20k')

# #aas
# #print(total_aasdis['tcrpeg'])
# Plot.aas_dis(aa2pos_dic,total_aasdis['tcrpeg'],total_aasdis['sonnia'],total_aasdis['tcrvae'],'AAs_20k')


# # # #vs,js
#Plot.v_dis(data_v,total_vsdis['tcrpeg'],total_vsdis['sonnia'],total_vsdis['tcrvae'],'vs_20k_whole')
# Plot.j_dis(data_j,total_jsdis['tcrpeg'],total_jsdis['sonnia'],total_jsdis['tcrvae'],'js_20k')

# #prob plot
# _,_,p_data,p_tcrpeg = eva.eva_prob('../../nlp_vae/data/pdf_test_sonnia.csv')
# print('done calculating')

# fig=plt.figure(figsize=(6,4),dpi=200)
# ax1=plt.subplot(111)
# ax1.set_ylim([-8,-1])
# ax1.set_xlim([-4.8,-1.5])
# ax1.locator_params(nbins=4)
# Plot.density_scatter(np.log10(p_data),np.log10(p_tcrpeg),bins = [10,50],ax=ax1,fig_name='test_prob')
















from tcrpeg.TCRpeg import TCRpeg
from tcrpeg.classification import classification
import pandas as pd
import numpy as np
import os
import argparse
from sklearn.model_selection import KFold
parser = argparse.ArgumentParser(description='')
parser.add_argument('--n',type=int,help='index of restart')
args = parser.parse_args()
#this is for tcrpeg

##train a model on positive data
# data = pd.read_csv('../../nlp_vae/data/sub_tasks/caTCR/normal_train_12_16.txt',names=['seq'])['seq'].values
# t_model = TCRpeg(hidden_size=128,num_layers = 3,load_data=True,embedding_path='data/embedding_32.txt',path_train=data,dropout=0.1)
# t_model.create_model()
# t_model.train_tcrpeg(40,16,5e-4,info='128_3l_40e_5e-4_16b_normal_train',record_dir='results/aug')

# t_model = TCRpeg(hidden_size=128,num_layers = 3,load_data=False,embedding_path='data/embedding_32.txt',dropout=0.1)
# t_model.create_model(load=True,path='results/aug/14:32:04/model.pth')
#num_aug = [10000,20000,30000,40000,50000,75000,100000]
num_aug = [45000,50000]
#num_aug = [8000,10000,15000,20000]
#num_aug = [n*2 for n in num_aug]
#num_aug = [12000]
# e_model = TCRpeg(hidden_size=512,num_layers = 3,load_data=False,embedding_path='data/embedding_32.txt',dropout=0.1)
# e_model.create_model(load=True,path='results/caTCR/23:02:07/model.pth')
tumor_train = pd.read_csv('data/classification/tumor_train_caTCR.txt',names=['seq'])['seq'].values
tumor_test = pd.read_csv('data/classification/tumor_test_caTCR.txt',names=['seq'])['seq'].values
tumor_whole = list(tumor_train) + list(tumor_test)
normal_train = pd.read_csv('data/classification/normal_train_caTCR.txt',names=['seq'])['seq'].values
normal_test = pd.read_csv('data/classification/normal_test_caTCR.txt',names=['seq'])['seq'].values
normal_whole = list(normal_train) + list(normal_test)
whole = list(tumor_whole) + list(normal_whole)
y_whole = [1] * len(tumor_whole) + [0] * len(normal_whole)
kf = KFold(n_splits=5,random_state=0,shuffle=True)
xx = list(range(len(whole)))
whole,y_whole = np.array(whole),np.array(y_whole)
kkk = 0
for train_idx,test_idx in kf.split(xx):
    x_train,x_test = whole[train_idx], whole[test_idx]
    y_train,y_test = y_whole[train_idx], y_whole[test_idx]
    tumor_train_idx,normal_train_idx = [k for k in range(len(y_train)) if y_train[k]==1],[k for k in range(len(y_train)) if y_train[k]==0]
    tumor_train,normal_train = x_train[tumor_train_idx],x_train[normal_train_idx]
    t_model = TCRpeg(hidden_size=128,num_layers = 3,load_data=True,embedding_path='data/embedding_32.txt',path_train=tumor_train,dropout=0.1)
    t_model.create_model()
    t_model.train_tcrpeg(40,16,5e-4)
    t_model2 = TCRpeg(hidden_size=128,num_layers = 3,load_data=True,embedding_path='data/embedding_32.txt',path_train=normal_train,dropout=0.1)
    t_model2.create_model()
    t_model2.train_tcrpeg(40,16,5e-4)
    #used the former 0.6 in augmentation
    for num_gens in num_aug:
        gens = t_model.generate_tcrpeg(num_gens,1000)
        gens = [x for x in gens if classification.valid_seq(x)]
        gens2 = t_model2.generate_tcrpeg(num_gens,1000)
        gens2 = [x for x in gens2 if classification.valid_seq(x)]
        prob_a = t_model.sampling_tcrpeg_batch(gens,2000)
        sorted_index = np.argsort(-prob_a)
        a_list_sorted = np.array(gens)[sorted_index]
        gens = a_list_sorted[int(len(a_list_sorted) * 0.6):]

        prob_a2 = t_model2.sampling_tcrpeg_batch(gens2,2000)
        sorted_index2 = np.argsort(-prob_a2)
        a_list_sorted2 = np.array(gens2)[sorted_index2]
        gens2 = a_list_sorted2[int(len(a_list_sorted2) * 0.6):]

        x_train  = list(x_train) + list(gens) + list(gens2)
        y_train = list(y_train) + [1] * len(gens) + [0] * len(gens2)

        # x_test = list(tumor_test) + list(normal_test)
        # y_test = [1] * len(tumor_test) + [0]*len(normal_test)
        e_model = TCRpeg(hidden_size=512,num_layers = 3,load_data=True,embedding_path='data/embedding_32.txt',path_train=x_train,dropout=0.1)
        e_model.create_model()
        e_model.train_tcrpeg(40,16,5e-4,info='128_3l_40e_5e-4_16b_tumor_train_aug')
        tcrpeg_c = classification(e_model,512*3,dropout=0.2,model_size='large')
        epoch = 30 
        # tcrpeg_c.train(x_train,y_train,epoch,8,1e-3,val_split=0,save_name='aug/caTCR_large_{}'.format(str(num_gens)))
        tcrpeg_c.train(x_train,y_train,epoch,8,1e-3,val_split=0)
        
        auc,aup,y_pres,y_trues = tcrpeg_c.evaluate(x_test,y_test,100,record_path='results/aug2/aug_cross_last_{}.txt'.format(num_gens))
        with open('results/aug2/predictions/{}_pres_large_cross_last_{}.txt'.format(str(kkk),str(num_gens)),'w') as f:
            for i in range(len(y_pres)):
                f.write(str(y_pres[i])+','+str(y_trues[i]) + '\n')
    kkk += 1

# from tcrpeg.TCRpeg import TCRpeg
# from tcrpeg.classification import classification
# import pandas as pd
# import numpy as np
# import os
# import argparse
# from sklearn.model_selection import KFold
# parser = argparse.ArgumentParser(description='')
# parser.add_argument('--n',type=int,help='index of restart')
# args = parser.parse_args()
# #this is for tcrpeg

# ##train a model on positive data
# # data = pd.read_csv('../../nlp_vae/data/sub_tasks/caTCR/normal_train_12_16.txt',names=['seq'])['seq'].values
# # t_model = TCRpeg(hidden_size=128,num_layers = 3,load_data=True,embedding_path='data/embedding_32.txt',path_train=data,dropout=0.1)
# # t_model.create_model()
# # t_model.train_tcrpeg(40,16,5e-4,info='128_3l_40e_5e-4_16b_normal_train',record_dir='results/aug')

# # t_model = TCRpeg(hidden_size=128,num_layers = 3,load_data=False,embedding_path='data/embedding_32.txt',dropout=0.1)
# # t_model.create_model(load=True,path='results/aug/14:32:04/model.pth')
# #num_aug = [10000,20000,30000,40000,50000,75000,100000]
# num_aug = [2000,4000,6000,8000,10000,15000,20000]
# #num_aug = [12000]
# # e_model = TCRpeg(hidden_size=512,num_layers = 3,load_data=False,embedding_path='data/embedding_32.txt',dropout=0.1)
# # e_model.create_model(load=True,path='results/caTCR/23:02:07/model.pth')
# tumor_train = pd.read_csv('data/classification/tumor_train_caTCR.txt',names=['seq'])['seq'].values
# tumor_test = pd.read_csv('data/classification/tumor_test_caTCR.txt',names=['seq'])['seq'].values
# tumor_whole = list(tumor_train) + list(tumor_test)
# normal_train = pd.read_csv('data/classification/normal_train_caTCR.txt',names=['seq'])['seq'].values
# normal_test = pd.read_csv('data/classification/normal_test_caTCR.txt',names=['seq'])['seq'].values
# normal_whole = list(normal_train) + list(normal_test)
# whole = list(tumor_whole) + list(normal_whole)
# y_whole = [1] * len(tumor_whole) + [0] * len(normal_whole)
# kf = KFold(n_splits=5,random_state=0,shuffle=True)
# xx = list(range(len(whole)))
# whole,y_whole = np.array(whole),np.array(y_whole)
# kkk = 0
# for train_idx,test_idx in kf.split(xx):
#     if kkk >=3:
#         break
#     x_train,x_test = whole[train_idx], whole[test_idx]
#     y_train,y_test = y_whole[train_idx], y_whole[test_idx]
#     tumor_train_idx,normal_train_idx = [k for k in range(len(y_train)) if y_train[k]==1],[k for k in range(len(y_train)) if y_train[k]==0]
#     tumor_train,normal_train = x_train[tumor_train_idx],x_train[normal_train_idx]
#     t_model = TCRpeg(hidden_size=128,num_layers = 3,load_data=True,embedding_path='data/embedding_32.txt',path_train=tumor_train,dropout=0.1)
#     t_model.create_model()
#     t_model.train_tcrpeg(40,16,5e-4)
#     t_model2 = TCRpeg(hidden_size=128,num_layers = 3,load_data=True,embedding_path='data/embedding_32.txt',path_train=normal_train,dropout=0.1)
#     t_model2.create_model()
#     t_model2.train_tcrpeg(40,16,5e-4)
#     for num_gens in num_aug:
#         gens = t_model.generate_tcrpeg(num_gens,1000)
#         gens = [x for x in gens if classification.valid_seq(x)]
#         #gens = gens[:(len(gens)//500)*500]
#         # prob_a = t_model.sampling_tcrpeg_batch(gens,2000)
#         # sorted_index = np.argsort(-prob_a)
#         # a_list_sorted = np.array(gens)[sorted_index]
#         # gens = a_list_sorted[:int(len(a_list_sorted) * 0.2)]
#         gens2 = t_model2.generate_tcrpeg(num_gens,1000)
#         gens2 = [x for x in gens2 if classification.valid_seq(x)]

#         x_train  = list(x_train) + list(gens) + list(gens2)
#         y_train = list(y_train) + [1] * len(gens) + [0] * len(gens2)

#         # x_test = list(tumor_test) + list(normal_test)
#         # y_test = [1] * len(tumor_test) + [0]*len(normal_test)
#         e_model = TCRpeg(hidden_size=512,num_layers = 3,load_data=True,embedding_path='data/embedding_32.txt',path_train=x_train,dropout=0.1)
#         e_model.create_model()
#         e_model.train_tcrpeg(40,16,5e-4,info='128_3l_40e_5e-4_16b_tumor_train_aug')
#         tcrpeg_c = classification(e_model,512*3,dropout=0.2,model_size='large')
#         epoch = 30 
#         # tcrpeg_c.train(x_train,y_train,epoch,8,1e-3,val_split=0,save_name='aug/caTCR_large_{}'.format(str(num_gens)))
#         tcrpeg_c.train(x_train,y_train,epoch,8,1e-3,val_split=0)
        
#         auc,aup,y_pres,y_trues = tcrpeg_c.evaluate(x_test,y_test,100,record_path='results/aug2/aug_cross_{}.txt'.format(num_gens))
#         with open('results/aug2/predictions/{}_pres_large_cross_{}.txt'.format(str(kkk),str(num_gens)),'w') as f:
#             for i in range(len(y_pres)):
#                 f.write(str(y_pres[i])+','+str(y_trues[i]) + '\n')
#     kkk += 1



from utils import *
import pandas as pd
import numpy as np
from collections import defaultdict
from matplotlib.lines import Line2D
from evaluate import evaluation
from tcrpeg.TCRpeg import TCRpeg
import os
from scipy.stats import pearsonr as pr

np.random.seed(0)
Plot = plotting()
model = TCRpeg(hidden_size=64,num_layers = 3,load_data=False,embedding_path='data/embedding_32.txt')
# path = '../../nlp_vae/results/models/11_22_2021/19:35:40/only_decoder_11_22_2021-19:35:40_40.pth'
path = 'results/models/15:07:30/model.pth'
model.create_model(load=True,path=path)

eva = evaluation(model=model,frac=0.1)

#prob plot
ks = []
_,_,p_data,p_tcrpeg = eva.eva_prob('../../nlp_vae/data/pdf_test_sonnia.csv',whole=False)
print('done calculating')

fig=plt.figure(figsize=(5,3),dpi=200)
ax1=plt.subplot(111)
ax1.set_ylim([-21,0])
ax1.set_xlim([-5.5,-2.5])
ax1.locator_params(nbins=4)
ax1.set_xlabel(r'$log_{10}P_{data}$')
ax1.set_ylabel(r'$log_{10}P_{infer}$')
ax1.plot([-5.5, -2.5], [-5.5, -2.5], color='k', linestyle='-', linewidth=2)
#Plot.density_scatter(np.log10(p_data),np.log10(p_tcrpeg),bins = [10,50],ax=ax1,fig_name='prob_tcrpeg',method='TCRpeg')
Plot.density_scatter(np.log10(p_data),np.log10(p_tcrpeg),bins = [10,50],ax=ax1,fig_name='prob_tcrpeg',method='TCRpeg')
ks.append(pr(p_data,p_tcrpeg)[0])
for i in range(4):
    _,_,p_data,p_tcrpeg = eva.eva_prob('../../nlp_vae/data/pdf_test_sonnia.csv',whole=False)
    ks.append(pr(p_data,p_tcrpeg)[0])
mean,std = round(np.mean(ks),4),round(np.std(ks),4)
#mean = 0.896
ax1.text(0.65, 0.32, r'r = %0.3f $\pm$ %0.3f' % (mean, std) , ha='center', va='center',transform = ax1.transAxes,size=10,color='k')
plt.tight_layout()
plt.savefig('results/pictures/prob_tcrpeg_20000.jpg',dpi=200)

# #sonnia
# ks = []
# d_sonnia = pd.read_csv('results/probs/pdf_test_sonnia_20000_0.csv')
# count,p_sonnia = d_sonnia['count'].values,d_sonnia['p_sonnia'].values
# p_sonnia += 1e-25 #to avoid 0
# p_data_sonnia,p_sonnia = count / np.sum(count), p_sonnia/np.sum(p_sonnia)
# print(kl_divergence(p_data_sonnia,p_sonnia))
# print(pr(p_data_sonnia,p_sonnia)[0])
# fig=plt.figure(figsize=(5,3),dpi=200)
# ax1=plt.subplot(111)
# ax1.set_ylim([-21,0])
# ax1.set_xlim([-5.5,-2.5])
# ax1.locator_params(nbins=4)
# ax1.set_xlabel(r'$log_{10}P_{data}$')
# ax1.set_ylabel(r'$log_{10}P_{infer}$')
# ax1.plot([-5.5, -2.5], [-5.5, -2.5], color='k', linestyle='-', linewidth=2)
# Plot.density_scatter(np.log10(p_data_sonnia),np.log10(p_sonnia) ,bins = [10,50],ax=ax1,method='soNNia')
# ks.append(pr(p_data_sonnia,p_sonnia)[0])
# for i in range(4):
#     d_sonnia = pd.read_csv('results/probs/pdf_test_sonnia_{}.csv'.format(i+1))
#     count,p_sonnia = d_sonnia['count'].values,d_sonnia['p_sonnia'].values
#     p_sonnia += 1e-25 #to avoid 0
#     p_data_sonnia,p_sonnia = count / np.sum(count), p_sonnia/np.sum(p_sonnia)
#     ks.append(pr(p_data_sonnia,p_sonnia)[0])
# mean,std = round(np.mean(ks),4),round(np.std(ks),4)
# ax1.text(0.65, 0.32, r'r = %0.3f $\pm$ %0.3f' % (mean, std) , ha='center', va='center',transform = ax1.transAxes,size=10,color='k')
# plt.tight_layout()
# plt.savefig('results/pictures/prob_sonnia_200000.jpg',dpi=200)

# #tcrvae
# ks = []
# d_vae = pd.read_csv('results/probs/pdf_test_tcrvae_20k_0.csv')
# count,p_tcrvae = d_vae['count'].values,d_vae['p_tcrvae'].values
# p_tcrvae += 1e-25 #to avoid 0
# p_data_tcrvae,p_tcrvae = count / np.sum(count), p_tcrvae/np.sum(p_tcrvae)
# print(kl_divergence(p_data_tcrvae,p_tcrvae))
# ks.append(pr(p_data_tcrvae,p_tcrvae)[0])
# print(pr(p_data_tcrvae,p_tcrvae)[0])
# fig=plt.figure(figsize=(5,3),dpi=200)
# ax1=plt.subplot(111)
# ax1.set_ylim([-21,0])
# ax1.set_xlim([-5.5,-2.5])
# ax1.locator_params(nbins=4)
# ax1.plot([-5.5, -2.5], [-5.5, -2.5], color='k', linestyle='-', linewidth=2)
# ax1.set_xlabel(r'$log_{10}P_{data}$')
# ax1.set_ylabel(r'$log_{10}P_{infer}$')
# Plot.density_scatter(np.log10(p_data_tcrvae),np.log10(p_tcrvae),bins = [10,50],ax=ax1,method='TCRvae')
# for i in range(4):
#     d_vae = pd.read_csv('results/probs/pdf_test_tcrvae_{}.csv'.format(i+1))
#     count,p_tcrvae = d_vae['count'].values,d_vae['p_tcrvae'].values
#     p_tcrvae += 1e-25 #to avoid 0
#     p_data_tcrvae,p_tcrvae = count / np.sum(count), p_tcrvae/np.sum(p_tcrvae)
#     ks.append(pr(p_data_tcrvae,p_tcrvae)[0])
# mean,std = round(np.mean(ks),4),round(np.std(ks),4)
# ax1.text(0.65, 0.32, r'r = %0.3f $\pm$ %0.3f' % (mean, std) , ha='center', va='center',transform = ax1.transAxes,size=10,color='k')
# plt.tight_layout()
# plt.savefig('results/pictures/prob_tcrvae_200000.jpg',dpi=200)


from setuptools import setup, find_packages

data_files_to_include = [('', ['README.md', 'LICENSE'])]

setup(name='tcrpeg',
      version='1.0.2',
      description='Recovering the repertoire probability, generating and encoding CDR3 sequences',
      long_description='To be added',
      url='https://github.com/jiangdada1221/TCRpeg',
      author='Yuepeng Jiang',
      author_email='jiangdada12344321@gmail.com',
      license='GPLv3',
      classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'Intended Audience :: Healthcare Industry',
            'Intended Audience :: Science/Research',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
            'Topic :: Scientific/Engineering :: Medical Science Apps.',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Natural Language :: English',
            'Programming Language :: Python :: 3',
            ],
      packages=find_packages(),
      install_requires=['numpy','torch','matplotlib','tqdm','pandas','scikit-learn','Scipy'],
      data_files = data_files_to_include,
      include_package_data=True,
      zip_safe=False)



from tcrpeg.TCRpeg import TCRpeg
from tcrpeg.classification import classification
import pandas as pd
import numpy as np
import os
from sklearn.metrics import roc_auc_score

# data = pd.read_csv('../../nlp_vae/data/sub_tasks/caTCR/whole_seqs_train_caTCR.txt',names=['seq'])['seq'].values
# t_model = TCRpeg(hidden_size=512,num_layers = 3,load_data=True,embedding_path='data/embedding_32.txt',path_train=data,dropout=0.1)
# t_model.create_model()
# t_model.train_tcrpeg(40,16,5e-4,info='512_3l_40e_5e-4_16b_whole_train_caTCR',record_dir='results/caTCR')
t_model = TCRpeg(hidden_size=512,num_layers = 3,load_data=False,embedding_path='data/embedding_32.txt',dropout=0.1)
t_model.create_model(load=True,path='results/caTCR/23:02:07/model.pth')

for iteration in range(5):
    tcrpeg_c = classification(t_model,512*3,dropout=0.2,model_size='large')
    tumor_train = pd.read_csv('data/classification/tumor_train_caTCR.txt',names=['seq'])['seq'].values
    tumor_test = pd.read_csv('data/classification/tumor_test_caTCR.txt',names=['seq'])['seq'].values
    normal_train = pd.read_csv('data/classification/normal_train_caTCR.txt',names=['seq'])['seq'].values
    normal_test = pd.read_csv('data/classification/normal_test_caTCR.txt',names=['seq'])['seq'].values
    x_train = list(tumor_train) + list(normal_train)
    y_train = [1]*len(tumor_train) + [0] * len(normal_train)
    x_test = list(tumor_test) + list(normal_test)
    y_test = [1] * len(tumor_test) + [0]*len(normal_test)
    tcrpeg_c.train(x_train,y_train,30,8,1e-3,val_split=0,save_name='caTCR/caTCR_large_{}'.format(str(iteration)))
    auc,aup,y_pres,y_trues = tcrpeg_c.evaluate(x_test,y_test,100,record_path='results/caTCR_large_2.txt')
    with open('results/caTCR/predictions/pres_large_2_{}.txt'.format(str(iteration)),'w') as f:
        for i in range(len(y_pres)):
            f.write(str(y_pres[i])+','+str(y_trues[i]) + '\n')
    _,_,y_pres,y_trues = tcrpeg_c.evaluate(x_train,y_train,100,record_path='results/caTCR_large_test_2.txt')
# for iteration in range(5):
#     tcrpeg_c = classification(t_model,512*3,dropout=0.2,model_size='medium')
#     tumor_train = pd.read_csv('data/classification/tumor_train_caTCR.txt',names=['seq'])['seq'].values
#     tumor_test = pd.read_csv('data/classification/tumor_test_caTCR.txt',names=['seq'])['seq'].values
#     normal_train = pd.read_csv('data/classification/normal_train_caTCR.txt',names=['seq'])['seq'].values
#     normal_test = pd.read_csv('data/classification/normal_test_caTCR.txt',names=['seq'])['seq'].values
#     x_train = list(tumor_train) + list(normal_train)
#     y_train = [1]*len(tumor_train) + [0] * len(normal_train)
#     x_test = list(tumor_test) + list(normal_test)
#     y_test = [1] * len(tumor_test) + [0]*len(normal_test)
#     tcrpeg_c.train(x_train,y_train,30,8,1e-3,save_name='caTCR/caTCR_{}'.format(str(iteration)))
#     auc,aup,y_pres,y_trues = tcrpeg_c.evaluate(x_test,y_test,100,record_path='results/caTCR_testing.txt')
#     with open('results/caTCR/predictions/pres_{}.txt'.format(str(iteration)),'w') as f:
#         for i in range(len(y_pres)):
#             f.write(str(y_pres[i])+','+str(y_trues[i]) + '\n')




import torch
from torch.autograd import Variable
import numpy as np
import torch.functional as F
import torch.nn.functional as F
from tqdm import tqdm
import pandas as pd

def word2vec(path,epochs,batch_size,device,record_path=None,lr=0.0001,window_size=2,embedding_dims = 32):
    '''
    Training the word2vec for embedding of amino acids
    @path: path to the file that records the sequences
    @batch_size: batch_size    
    @device: GPU or CPU; 'cuda:0'-GPU, 'cpu'-CPU
    @record_path: path to record the embedding for each AA; should ended with '.txt'
    @lr: initial learning rate
    @window_size: the window_size used for constructing training pairs (see word2vec paper for details)

    #return: the embedding (a numpy array)
    '''
    def get_input_layer(word_idxs):
        #x = torch.zeros(vocabulary_size).float()
        x = torch.zeros((len(word_idxs),vocabulary_size)).float()
        x[list(range(len(word_idxs))),word_idxs] = 1.0  # n x emb
        assert torch.sum(x) == len(word_idxs)
        return x #one hot vector
    #assert record_path.endswith('.txt')

    #list of a string is served as default tokenizer
    aas = 'sACDEFGHIKLMNPQRSTVWYe'
    aa2idx = {aas[i]:i for i in range(len(aas))}
    idx2aa = {v: k for k, v in aa2idx.items()}
    vocabulary_size = len(aas)
    batch_size = batch_size
    device = device    
    if isinstance(path,list) or isinstance(path,np.ndarray):
        tcr_sequences = np.array(path)
    else :
        tcr_sequences = pd.read_csv(path,sep='\t')['seq'].values

    tokenized_seqs = [['s']+list(seq)+['e'] for seq in tcr_sequences]

    idx_pairs = []
    # for each sentence
    for seq in tokenized_seqs:
        indices = [aa2idx[word] for word in seq]
        # for each word, threated as center word
        for center_word_pos in range(len(indices)):
            # for each window position
            for w in range(-window_size, window_size + 1):
                context_word_pos = center_word_pos + w
                # make soure not jump out sentence
                if context_word_pos < 0 or context_word_pos >= len(indices) or center_word_pos == context_word_pos:
                    continue
                context_word_idx = indices[context_word_pos]
                idx_pairs.append((indices[center_word_pos], context_word_idx))

    idx_pairs = np.array(idx_pairs) # N x 2
    W1 = Variable(torch.randn(embedding_dims, vocabulary_size,device=device).float(), requires_grad=True) 
    W2 = Variable(torch.randn(vocabulary_size, embedding_dims,device=device).float(), requires_grad=True) 
    num_epochs = epochs
    learning_rate = lr

    for epo in range(num_epochs):
        loss_val = 0        
        iter = len(idx_pairs) // batch_size 
        for iter in tqdm(range(iter)):
            #x = Variable(get_input_layer(data)).float()
            x = Variable(get_input_layer(idx_pairs[iter*batch_size:(iter+1)*batch_size,0])).float().to(device)  #nxv
        
            y_true = Variable(torch.from_numpy(np.array(idx_pairs[iter*batch_size:(iter+1)*batch_size,1])).long()).to(device)
            
            z1 = torch.matmul(W1, torch.transpose(x,0,1))
            z2 = torch.matmul(W2, z1) # |v| x batch_size
            
            log_softmax = F.log_softmax(z2, dim=0).permute([1,0]) #batch_size x |v|
            loss = F.nll_loss(log_softmax, y_true)
            loss_val += loss.item()
            loss.backward()
            W1.data -= learning_rate * W1.grad.data
            W2.data -= learning_rate * W2.grad.data

            W1.grad.data.zero_()
            W2.grad.data.zero_()
            
        if epo % 2 == 0:    
            print(f'Loss at epo {epo+1}: {loss_val * 32 /len(idx_pairs)}')
        if epo % 40 == 0:
            learning_rate *= 0.2

    embedding = W1.cpu().data.numpy()
    if record_path is not None:
        with open(record_path, 'w') as f:
            for i in range(22):
                f.write(aas[i]+',')
                for j in range(embedding_dims):
                    f.write(str(embedding[j,i]))
                    if j != embedding_dims-1:
                        f.write(',')
                f.write('\n')
    return embedding





import torch
import torch.nn as nn 
from torch.nn.utils.rnn import pack_padded_sequence,pad_packed_sequence
from tqdm import tqdm
import numpy as np

#note that, the padding_idx is 0, the idx for START token is 1
#the idx for AAs start from 2 and the idx for END stoken is 22


class TCRpeg_model(nn.Module):
    def __init__(self,embedding_layer,embedding_size,hidden_size,dropout,max_length=30,sos_idx=1,pad_idx=0,eos_idx=22,device='cuda:0',num_layers=1):
        super(TCRpeg_model,self).__init__()
        self.max_length = max_length
        self.sos_idx=sos_idx
        self.eos_idx=  eos_idx
        self.pad_idx = pad_idx
        # self.embedding_dropout= embedding_dropout
        self.embedding = embedding_layer
        self.embedding_size=  embedding_size
        self.hidden_size=hidden_size
        
        self.decoder_rnn = nn.GRU(embedding_size,hidden_size,num_layers,dropout=dropout)
        self.num_layers = num_layers
        self.out_layer=  nn.Linear(hidden_size, 22)
        self.device = device
    
    def forward(self,seqs,length,need_hidden=False):
        #note that, for the RNN layer in torch, the batch_num is not 
        #at the first dimension (batch_first=False), which results in those permute' below
        self.batch_size = seqs.size(0)
        sorted_lengths, sorted_idx = torch.sort(length,descending=True) #sorting is used for pack_padded_sequence
      
        seqs = seqs[sorted_idx]
      
        input_emb = self.embedding(seqs).to(torch.float32).permute([1,0,2]) 
   
        packed_input = pack_padded_sequence(input_emb,sorted_lengths.data.tolist())

        outputs,h_n= self.decoder_rnn(packed_input)
        
        padded_outputs = pad_packed_sequence(outputs)[0].permute([1,0,2]) 
        h_n = h_n.permute([1,0,2])

        _, reversed_idx = torch.sort(sorted_idx)
        padded_outputs = padded_outputs[reversed_idx]
        h_n = h_n[reversed_idx] 
    
        logp = nn.functional.log_softmax(self.out_layer(padded_outputs),dim=-1) # batch_size x max_length x 22
        # print(length)
        # print(logp[:,length-1,:])
        # print(logp[:,length-1,:].size())
        # exit()
        #logp = logp.permute([0,2,1]) #batch_size x 22 x max_length
        if need_hidden:
            return logp,h_n
        else :
            return logp

class TCRpeg_vj_model(nn.Module):
    def __init__(self,embedding_layer,embedding_size,hidden_size,dropout,max_length=30,sos_idx=1,pad_idx=0,eos_idx=22,device='cuda:0',num_layers=1,num_v=43,num_j=12):
        super(TCRpeg_vj_model,self).__init__()
        self.max_length = max_length
        self.sos_idx=sos_idx
        self.eos_idx=  eos_idx
        self.pad_idx = pad_idx
        # self.embedding_dropout= embedding_dropout
        self.embedding = embedding_layer
        self.embedding_size=  embedding_size
        self.hidden_size=hidden_size
        self.decoder_rnn = nn.GRU(embedding_size,hidden_size,num_layers,dropout=dropout)
        self.num_layers = num_layers
        self.out_layer=  nn.Linear(hidden_size, 22)
        self.device = device
        #self.v_layer = nn.Sequential(nn.Linear(hidden_size * num_layers,128),nn.ReLU(),nn.Linear(128,num_v)) #try a simple one first
        #self.j_layer = nn.Sequential(nn.Linear(hidden_size * num_layers,64),nn.ReLU(),nn.Linear(64,num_j))
        self.v_layer = nn.Sequential(nn.Linear(hidden_size * num_layers,num_v)) #try a simple one first
        self.j_layer = nn.Sequential(nn.Linear(hidden_size * num_layers,num_j))
    
    def forward(self,seqs,length,need_hidden=False):
        self.batch_size = seqs.size(0)
        sorted_lengths, sorted_idx = torch.sort(length,descending=True)


        seqs = seqs[sorted_idx]
    
        input_emb = self.embedding(seqs).to(torch.float32).permute([1,0,2])
   
        packed_input = pack_padded_sequence(input_emb,sorted_lengths.data.tolist())

        outputs,h_n= self.decoder_rnn(packed_input)
        #h_n is L x B x Hout
      

        padded_outputs = pad_packed_sequence(outputs)[0].permute([1,0,2]) 
        h_n = h_n.permute([1,0,2]) #B x layer X Hout
        #print(padded_outputs.size()) #batch_size x max_length x hidden_size
        #padded_outputs = padded_outputs.contiguous() # check this 
        _, reversed_idx = torch.sort(sorted_idx)
        padded_outputs = padded_outputs[reversed_idx]
        h_n = h_n[reversed_idx] 
    
        logp = nn.functional.log_softmax(self.out_layer(padded_outputs),dim=-1) # batch_size x max_length x 22

        #logp = logp.permute([0,2,1]) #batch_size x 22 x max_length
        h_n = h_n.view(self.batch_size,-1)
    
        v_pre = self.v_layer(h_n) #check the size of h_n
        j_pre = self.j_layer(h_n)
        if need_hidden:
            return logp,v_pre,j_pre,h_n
        else :
            return logp,v_pre,j_pre

#Models below are used for classification task (to show the usefulness of the embedding of tcrpeg)
class FC_NN_small(nn.Module):
    def __init__(self,embedding_size,last_layer=True,device='cuda:0'):
        #the model should be a vae_nlp model
        super(FC_NN_small,self).__init__()
        #self.vae_model= model
        #self.vae_model.model.eval()
        self.embeeding_size=  embedding_size
        self.device= device
        self.last_layer = last_layer #whether to use the last layer of GRU hidden units
        self.f1 = nn.Linear(embedding_size,1)
        self.embedding_size = embedding_size
        self.sigmoid = nn.Sigmoid()
    
    def forward(self,embedding):
        return self.sigmoid(self.f1(embedding))


class FC_NN_medium(nn.Module):
    def __init__(self,embedding_size,dropout,last_layer=True,device='cuda:0',batch_norm=False):

        super(FC_NN_medium,self).__init__()
        self.embeeding_size=  embedding_size
        self.dropout = nn.Dropout(p=dropout)
        self.device= device
        self.last_layer = last_layer #whether to use the last layer of GRU hidden units
        self.f1 = nn.Linear(embedding_size,embedding_size//4)
        self.relu = nn.ReLU()
        self.f2 = nn.Linear(embedding_size // 4,1)
        self.embedding_size = embedding_size
        self.sigmoid = nn.Sigmoid()
        self.bn = nn.BatchNorm1d(embedding_size // 4)
        self.batch_norm = batch_norm
    
    def forward(self,embedding):
        if self.batch_norm:
            f1 = self.dropout(self.relu(self.bn(self.f1(embedding))))
        else :
            f1 = self.dropout(self.relu(self.f1(embedding)))
        f2 = self.sigmoid(self.f2(f1))
        return f2

class FC_NN_large(nn.Module):
    def __init__(self,embedding_size,dropout,last_layer=True,device='cuda:0',batch_norm=False):
        #the model should be a vae_nlp model
        super(FC_NN_large,self).__init__()
        #self.vae_model= model
        #self.vae_model.model.eval()
        self.embeeding_size=  embedding_size
        self.dropout = nn.Dropout(p=dropout)
        self.device= device
        self.last_layer = last_layer #whether to use the last layer of GRU hidden units
        self.f1 = nn.Linear(embedding_size,embedding_size // 4)
        self.f2 = nn.Linear(embedding_size // 4 , embedding_size // 16)
        self.f3 = nn.Linear(embedding_size // 16, 1)
        self.embedding_size = embedding_size
        self.sigmoid = nn.Sigmoid()
        self.relu = nn.ReLU()
        self.batch_norm = batch_norm
        self.bn1 = nn.BatchNorm1d(embedding_size//4)
        self.bn2 = nn.BatchNorm1d(embedding_size//16)
        
    
    def forward(self,embedding):
        if self.batch_norm:
            s1 = self.dropout(self.relu(self.bn1(self.f1(embedding))))
            s2 = self.dropout(self.relu(self.bn2(self.f2(s1))))
        else :
            s1 = self.dropout(self.relu(self.f1(embedding)))
            s2 = self.dropout(self.relu(self.f2(s1)))
        return self.sigmoid(self.f3(s2))



import pandas as pd
import numpy as np
from utils import *
import os
#need 22 x 5

# d1 = pd.read_csv('results/TCRGP/baseline.txt',names=['es','aucs','aups','none'])
# es,aups = d1['es'].values,d1['aups'].values
# a1 = []
# for i in range(5):
#     a1.append(list(aups[i*22:(i+1)*22]))

# d2 = pd.read_csv('results/TCRGP/tcrpeg_fc.txt',names=['es','aucs','aups','none'])
# aups = d2['aups'].values
# a2 = []
# for i in range(5):
#     a2.append(list(aups[i*22:(i+1)*22]))

# es = es[:22]
# a1,a2 = np.array(a1).T,np.array(a2).T
# Plot = plotting()
# df = Plot.plot_violin(a2,a1,list(es),'results/pictures/tcrgp')
# Plot.plot_sup(a2,a1,'results/pictures/sup_auc')

# #aug for baseline
# d1 = pd.read_csv('results/TCRGP/baseline.txt',names=['es','aucs','aups','none'])
# es,aups = d1['es'].values,d1['aups'].values
# a1 = []
# Plot = plotting()
# for i in range(5):
#     a1.append(list(aups[i*22:(i+1)*22]))
# a1 = np.array(a1).T
# nums = [20,40,60,80,100]
# es = es[:22]
# for num in nums:
#     a2 = []
#     d2 = pd.read_csv('results/TCRGP/aug/baseline/baseline_{}.txt'.format(num),names=['es','aucs','aups','none'])
#     aups = d2['aups'].values

#     for i in range(5):
#         a2.append(list(aups[i*22:(i+1)*22]))
#     a2 = np.array(a2).T
#     df = Plot.plot_violin(a2,a1,list(es),'results/pictures/tcrgp_baseline_{}'.format(num),method1='Augmentation Size - {}'.format(num),method2='Without Augmentation')
#     Plot.plot_sup(a2,a1,'results/pictures/sup_auc_baseline_{}'.format(num),xlabel='Without augmentation',ylabel='Augmentation size - {}'.format(num))

#aug for fc
d1 = pd.read_csv('results/TCRGP/tcrpeg_fc.txt',names=['es','aucs','aups','none'])
es,aups = d1['es'].values,d1['aups'].values
a1 = []
Plot = plotting()
for i in range(5):
    a1.append(list(aups[i*22:(i+1)*22]))
a1 = np.array(a1).T
nums = [20,40,60,80,100]
es = es[:22]
for num in nums:
    a2 = []
    d2 = pd.read_csv('results/TCRGP/aug/fc/tcrpeg_fc_{}.txt'.format(num),names=['es','aucs','aups','none'])
    aups = d2['aups'].values

    for i in range(5):
        a2.append(list(aups[i*22:(i+1)*22]))
    a2 = np.array(a2).T
    df = Plot.plot_violin(a2,a1,list(es),'results/pictures/tcrgp_fc_{}'.format(num),method1='Augmentation Size - {}'.format(num),method2='Without Augmentation')
    Plot.plot_sup(a2,a1,'results/pictures/sup_auc_fc_{}'.format(num),xlabel='Without augmentation',ylabel='Augmentation size - {}'.format(num))
